from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MIN, ChannelEvent, ParticipantInfoIndex
from raiden_contracts.tests.fixtures.channel import call_settle
from raiden_contracts.tests.utils import (
    LOCKSROOT_OF_NO_LOCKS,
    NONEXISTENT_LOCKSROOT,
    UINT256_MAX,
    ChannelValues,
    LockedAmounts,
    LockIndex,
    call_and_transact,
    fake_bytes,
    get_unlocked_amount,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.events import check_channel_unlocked
from raiden_contracts.utils.pending_transfers import (
    get_locked_amount,
    get_packed_transfers,
    get_pending_transfers_tree,
    random_secret,
)

# Account names like 'A', 'B', 'C' are intuitive here.
# pytest: disable=C0103


def test_packed_transfer_0_items(token_network_test_utils: Contract) -> None:
    """getHashAndUnlockedAmount() returns a reasonable return value for no items"""
    (
        locksroot,
        unlocked_amount,
    ) = token_network_test_utils.functions.getHashAndUnlockedAmountPublic(b"").call()
    assert locksroot == LOCKSROOT_OF_NO_LOCKS
    assert unlocked_amount == 0


def test_0_item_unlockable(web3: Web3, token_network_test_utils: Contract) -> None:
    """Test getHashRootAndUnlockedAmount() on zero items whose secret has been registered"""
    pending_transfers_tree = get_pending_transfers_tree(
        web3=web3, unlockable_amounts=[], expired_amounts=[]
    )

    (
        locksroot,
        unlocked_amount,
    ) = token_network_test_utils.functions.getHashAndUnlockedAmountPublic(
        pending_transfers_tree.packed_transfers
    ).call()

    total_hash = pending_transfers_tree.hash_of_packed_transfers
    assert locksroot == total_hash
    assert unlocked_amount == 0


def test_1_item_unlockable(
    web3: Web3,
    get_accounts: Callable,
    token_network_test_utils: Contract,
    secret_registry_contract: Contract,
) -> None:
    """Test getHashRootAndUnlockedAmount() on a single item whose secret has been registered"""
    A = get_accounts(1)[0]
    pending_transfers_tree = get_pending_transfers_tree(
        web3=web3, unlockable_amounts=[6], expired_amounts=[]
    )

    call_and_transact(
        secret_registry_contract.functions.registerSecret(
            pending_transfers_tree.unlockable[0][LockIndex.SECRET]
        ),
        {"from": A},
    )
    assert (
        secret_registry_contract.functions.getSecretRevealBlockHeight(
            pending_transfers_tree.unlockable[0][LockIndex.SECRETHASH]
        ).call()
        == web3.eth.block_number
    )

    (
        locksroot,
        unlocked_amount,
    ) = token_network_test_utils.functions.getHashAndUnlockedAmountPublic(
        pending_transfers_tree.packed_transfers
    ).call()

    total_hash = pending_transfers_tree.hash_of_packed_transfers
    assert locksroot == total_hash
    assert unlocked_amount == 6


def test_get_hash_length_fail(
    web3: Web3,
    get_accounts: Callable,
    token_network_test_utils: Contract,
    secret_registry_contract: Contract,
) -> None:
    """Test getHashAndUnlockedAmount() on inputs of irregular lengths"""
    network_utils = token_network_test_utils
    A = get_accounts(1)[0]
    pending_transfers_tree = get_pending_transfers_tree(web3, [2, 3, 6], [5])

    call_and_transact(
        secret_registry_contract.functions.registerSecret(
            pending_transfers_tree.unlockable[0][LockIndex.SECRET]
        ),
        {"from": A},
    )
    assert (
        secret_registry_contract.functions.getSecretRevealBlockHeight(
            pending_transfers_tree.unlockable[0][LockIndex.SECRETHASH]
        ).call()
        == web3.eth.block_number
    )

    packed = pending_transfers_tree.packed_transfers

    # packed length must be a multiple of 96
    with pytest.raises(TransactionFailed, match="TN: invalid locks size"):
        network_utils.functions.getHashAndUnlockedAmountPublic(packed[0:-1]).call()
    # last lock only contains expiration + locked_amount
    with pytest.raises(TransactionFailed, match="TN: invalid locks size"):
        network_utils.functions.getHashAndUnlockedAmountPublic(packed[0:-32]).call()
    # last lock only contains expiration
    with pytest.raises(TransactionFailed, match="TN: invalid locks size"):
        network_utils.functions.getHashAndUnlockedAmountPublic(packed[0:-64]).call()

    assert len(packed) % 96 == 0
    network_utils.functions.getHashAndUnlockedAmountPublic(packed).call()
    network_utils.functions.getHashAndUnlockedAmountPublic(packed[0:-96]).call()


def test_odd_even_components(
    web3: Web3,
    get_accounts: Callable,
    token_network_test_utils: Contract,
    reveal_secrets: Callable,
) -> None:
    """Test getHashAndUnlockedAmount() on an odd/even number of locks"""
    A = get_accounts(1)[0]

    # Even number of locks
    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 8, 3])
    reveal_secrets(A, pending_transfers_tree.unlockable)

    (
        locksroot,
        unlocked_amount,
    ) = token_network_test_utils.functions.getHashAndUnlockedAmountPublic(
        pending_transfers_tree.packed_transfers
    ).call()
    total_hash = pending_transfers_tree.hash_of_packed_transfers

    assert locksroot == total_hash
    assert unlocked_amount == 9

    # Odd number of locks
    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 8])
    reveal_secrets(A, pending_transfers_tree.unlockable)

    (
        locksroot,
        unlocked_amount,
    ) = token_network_test_utils.functions.getHashAndUnlockedAmountPublic(
        pending_transfers_tree.packed_transfers
    ).call()
    total_hash = pending_transfers_tree.hash_of_packed_transfers

    assert locksroot == total_hash
    assert unlocked_amount == 9


def test_locks_order(
    web3: Web3,
    get_accounts: Callable,
    token_network_test_utils: Contract,
    reveal_secrets: Callable,
    token_network: Contract,
    create_settled_channel: Callable,
) -> None:
    """Shuffling the leaves usually changes the root, but sometimes not"""
    network_utils = token_network_test_utils
    (A, B) = get_accounts(2)
    types = ["uint256", "uint256", "bytes32"]

    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 8, 3])
    reveal_secrets(A, pending_transfers_tree.unlockable)

    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree.locked_amount,
        pending_transfers_tree.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
    )

    # Pending locks are in the insertion order.
    # If we change the order, we change the computed hash.
    wrong_order = pending_transfers_tree.transfers
    wrong_order[1], wrong_order[0] = wrong_order[0], wrong_order[1]
    wrong_order_packed = get_packed_transfers(wrong_order, types)
    (
        locksroot,
        unlocked_amount,
    ) = network_utils.functions.getHashAndUnlockedAmountPublic(wrong_order_packed).call()
    # If we change the order, we change the computed hash.
    assert locksroot != pending_transfers_tree.hash_of_packed_transfers
    assert unlocked_amount == 9
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(channel_identifier, B, A, wrong_order_packed).call()

    wrong_order = pending_transfers_tree.transfers
    wrong_order[2], wrong_order[0] = wrong_order[0], wrong_order[2]
    wrong_order_packed = get_packed_transfers(wrong_order, types)
    (
        locksroot,
        unlocked_amount,
    ) = network_utils.functions.getHashAndUnlockedAmountPublic(wrong_order_packed).call()
    assert locksroot != pending_transfers_tree.hash_of_packed_transfers
    assert unlocked_amount == 9
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(channel_identifier, B, A, wrong_order_packed).call()

    wrong_order = pending_transfers_tree.transfers
    wrong_order[0], wrong_order[-1] = wrong_order[-1], wrong_order[0]
    wrong_order_packed = get_packed_transfers(wrong_order, types)
    (
        locksroot,
        unlocked_amount,
    ) = network_utils.functions.getHashAndUnlockedAmountPublic(wrong_order_packed).call()
    assert locksroot != pending_transfers_tree.hash_of_packed_transfers
    assert unlocked_amount == 9
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(channel_identifier, B, A, wrong_order_packed).call()

    (locksroot, unlocked_amount,) = network_utils.functions.getHashAndUnlockedAmountPublic(
        pending_transfers_tree.packed_transfers
    ).call()
    assert locksroot == pending_transfers_tree.hash_of_packed_transfers
    assert unlocked_amount == 9
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree.packed_transfers
        )
    )


def test_lock_data_from_packed_locks(
    web3: Web3,
    get_accounts: Callable,
    token_network_test_utils: Contract,
    secret_registry_contract: Contract,
    reveal_secrets: Callable,
) -> None:
    """Test getLockDataFromLockPublic() on various offsets"""
    network_utils = token_network_test_utils
    A = get_accounts(1)[0]

    unlockable_amounts = [3, 5]
    expired_amounts = [2, 8, 7]
    pending_transfers_tree = get_pending_transfers_tree(
        web3, unlockable_amounts, expired_amounts, max_expiration_delta=5
    )
    reveal_secrets(A, pending_transfers_tree.unlockable)

    def claimable(index: int) -> int:
        amount = pending_transfers_tree.transfers[index][1]
        return amount if amount in unlockable_amounts else 0

    # Lock data is ordered lexicographically, regardless of expiration status
    claimable_amount = network_utils.functions.getLockedAmountFromLockPublic(
        pending_transfers_tree.packed_transfers, 32
    ).call()
    assert claimable_amount == claimable(0)

    claimable_amount = network_utils.functions.getLockedAmountFromLockPublic(
        pending_transfers_tree.packed_transfers, 32 + 96
    ).call()
    assert claimable_amount == claimable(1)

    claimable_amount = network_utils.functions.getLockedAmountFromLockPublic(
        pending_transfers_tree.packed_transfers, 32 + 2 * 96
    ).call()
    assert claimable_amount == claimable(2)

    claimable_amount = network_utils.functions.getLockedAmountFromLockPublic(
        pending_transfers_tree.packed_transfers, 32 + 3 * 96
    ).call()
    assert claimable_amount == claimable(3)

    claimable_amount = network_utils.functions.getLockedAmountFromLockPublic(
        pending_transfers_tree.packed_transfers, 32 + 4 * 96
    ).call()
    assert claimable_amount == claimable(4)

    # Register last secret after expiration
    mine_blocks(web3, 5)
    last_lock = pending_transfers_tree.expired[-1]
    # expiration
    assert web3.eth.block_number > last_lock[0]
    # register secret
    call_and_transact(secret_registry_contract.functions.registerSecret(last_lock[3]))
    # ensure registration was done
    assert (
        secret_registry_contract.functions.getSecretRevealBlockHeight(last_lock[2]).call()
        == web3.eth.block_number
    )

    # Check that last secret is still regarded as expired
    claimable_amount = network_utils.functions.getLockedAmountFromLockPublic(
        pending_transfers_tree.packed_transfers, 32 + 4 * 96
    ).call()
    assert claimable_amount == claimable(4)

    # If the offset is bigger than the length of the packed locks, return (0, 0)
    claimable_amount = network_utils.functions.getLockedAmountFromLockPublic(
        pending_transfers_tree.packed_transfers, 32 + 5 * 96
    ).call()
    assert claimable_amount == 0


def test_unlock_wrong_locksroot(
    web3: Web3, token_network: Contract, create_settled_channel: Callable, get_accounts: Callable
) -> None:
    """Test unlocking with wrong pending locks"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    pending_transfers_tree_A = get_pending_transfers_tree(web3, [1, 3, 5], [], settle_timeout)
    pending_transfers_tree_A_fake = get_pending_transfers_tree(web3, [1, 3, 6], [], settle_timeout)

    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree_A.locked_amount,
        pending_transfers_tree_A.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A_fake.packed_transfers
        ).call()

    # Fails for an empty packed locks
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(channel_identifier, B, A, b"").call()

    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        )
    )


def test_channel_unlock_bigger_locked_amount(
    web3: Web3,
    token_network: Contract,
    custom_token: Contract,
    secret_registry_contract: Contract,
    create_settled_channel: Callable,
    get_accounts: Callable,
    reveal_secrets: Callable,
) -> None:
    """Test an unlock() call that claims too little tokens"

    When an unlock() call does not contain enough pending locks to claim
    the locked amount declared in the settleChannel() call, the difference goes
    to the other party.
    """
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree_A = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_A.unlockable)
    unlocked_amount = get_unlocked_amount(
        secret_registry_contract, pending_transfers_tree_A.packed_transfers
    )

    # We settle the channel with a bigger locked amount than we will need for the
    # actual pending transfers
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree_A.locked_amount + 1,
        pending_transfers_tree_A.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_contract == pending_transfers_tree_A.locked_amount + 1

    # This should pass, even though the locked amount in storage is bigger. The rest of the
    # tokens is sent to A, as tokens corresponding to the locks that could not be unlocked.
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        )
    )
    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_A == (
        pre_balance_A + pending_transfers_tree_A.locked_amount - unlocked_amount + 1
    )
    assert balance_B == pre_balance_B + unlocked_amount
    assert balance_contract == 0


def test_channel_unlock_smaller_locked_amount(
    web3: Web3,
    token_network: Contract,
    custom_token: Contract,
    secret_registry_contract: Contract,
    create_settled_channel: Callable,
    get_accounts: Callable,
    reveal_secrets: Callable,
) -> None:
    """Test an unlock() call that claims too many tokens

    When settleChannel() call computes a smaller amount of locked tokens than
    the following unlock() call, the settleChannel() computation is stronger and
    the participant receives less tokens. Stealing tokens from other channels
    is then prevented."""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree_A = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_A.unlockable)
    unlocked_amount = get_unlocked_amount(
        secret_registry_contract, pending_transfers_tree_A.packed_transfers
    )

    # We settle the channel with a smaller locked amount than we will need for the
    # actual pending transfers
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree_A.locked_amount - 1,
        pending_transfers_tree_A.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_contract == pending_transfers_tree_A.locked_amount - 1

    # This should pass, even though the locked amount in storage is smaller.
    # B will receive less tokens.
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        )
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_A == (
        pre_balance_A + pending_transfers_tree_A.locked_amount - unlocked_amount - 1
    )
    assert balance_B == pre_balance_B + unlocked_amount
    assert balance_contract == 0


def test_channel_unlock_bigger_unlocked_amount(
    web3: Web3,
    token_network: Contract,
    custom_token: Contract,
    secret_registry_contract: Contract,
    create_settled_channel: Callable,
    get_accounts: Callable,
    reveal_secrets: Callable,
) -> None:
    """unlock() transfers not more than the locked amount for more expensive unlock() demands"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree_A = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_A.unlockable)
    unlocked_amount = get_unlocked_amount(
        secret_registry_contract, pending_transfers_tree_A.packed_transfers
    )
    assert unlocked_amount < pending_transfers_tree_A.locked_amount

    # We settle the channel with a smaller locked amount than the amount that can be unlocked
    channel_identifier = create_settled_channel(
        A,
        unlocked_amount - 1,
        pending_transfers_tree_A.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_contract == unlocked_amount - 1

    # This should pass, even though the locked amount in storage is smaller.
    # B will receive the entire locked amount, corresponding to the locks that have been unlocked
    # and A will receive nothing.
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        )
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_A == pre_balance_A
    assert balance_B == pre_balance_B + unlocked_amount - 1
    assert balance_contract == 0


def test_channel_unlock_no_locked_amount_fail(
    web3: Web3,
    token_network: Contract,
    create_settled_channel: Callable,
    get_accounts: Callable,
    reveal_secrets: Callable,
) -> None:
    """After settleChannel() is called with zero locked amount, unlock() calls fail"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree_A = get_pending_transfers_tree(web3, [2, 5], [4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_A.unlockable)

    channel_identifier = create_settled_channel(
        A, 0, LOCKSROOT_OF_NO_LOCKS, B, 0, LOCKSROOT_OF_NO_LOCKS, settle_timeout
    )

    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(channel_identifier, B, A, b"").call()
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        ).call()


def test_channel_unlock(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    close_and_update_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """unlock() on pending transfers with unlockable and expired locks should
    split the locked amount accordingly, to both parties"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    values_A = ChannelValues(deposit=20, transferred=5)
    values_B = ChannelValues(deposit=30, transferred=40)

    # Create channel and deposit
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, values_A.deposit, B)
    channel_deposit(channel_identifier, B, values_B.deposit, A)

    # Mock pending transfers data
    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    values_B.locksroot = pending_transfers_tree.hash_of_packed_transfers
    values_B.locked_amounts = LockedAmounts(
        claimable_locked=get_locked_amount(pending_transfers_tree.transfers)
    )

    # Reveal secrets before settlement window ends
    reveal_secrets(A, pending_transfers_tree.unlockable)

    close_and_update_channel(channel_identifier, A, values_A, B, values_B)

    # Settlement window must be over before settling the channel
    mine_blocks(web3, settle_timeout)

    call_settle(token_network, channel_identifier, A, values_A, B, values_B)

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    info_B = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert info_B[ParticipantInfoIndex.LOCKSROOT] == values_B.locksroot
    assert info_B[ParticipantInfoIndex.LOCKED_AMOUNT] == values_B.locked_amounts.locked

    # Unlock the tokens
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        )
    )

    info_B = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert info_B[ParticipantInfoIndex.LOCKSROOT] == NONEXISTENT_LOCKSROOT
    assert info_B[ParticipantInfoIndex.LOCKED_AMOUNT] == 0

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_A == pre_balance_A + 9
    assert balance_B == pre_balance_B + 6
    assert balance_contract == pre_balance_contract - values_B.locked_amounts.locked


@pytest.mark.slow
def test_channel_settle_and_unlock(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_settled_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """Regular channel life-cycle: open -> settle -> unlock -> open -> settle -> unlock"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree_1 = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_1.unlockable)

    # Settle the channel
    channel_identifier1 = create_settled_channel(
        A,
        pending_transfers_tree_1.locked_amount,
        pending_transfers_tree_1.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier1, B, A, pending_transfers_tree_1.packed_transfers
        ),
        {"from": A},
    )

    # Mock pending transfers data for a reopened channel
    pending_transfers_tree_2 = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_2.unlockable)

    # Settle the channel again
    channel_identifier2 = create_settled_channel(
        A,
        pending_transfers_tree_2.locked_amount,
        pending_transfers_tree_2.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    # 2nd unlocks should go through
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier2, B, A, pending_transfers_tree_2.packed_transfers
        ),
        {"from": A},
    )

    # Edge channel life-cycle: open -> settle -> open -> settle ->  unlock1 -> unlock2

    # Mock pending transfers data
    pending_transfers_tree_1 = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_1.unlockable)

    # Settle the channel
    channel_identifier3 = create_settled_channel(
        A,
        pending_transfers_tree_1.locked_amount,
        pending_transfers_tree_1.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    # Mock pending transfers data for a reopened channel
    pending_transfers_tree_2 = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_2.unlockable)

    # Settle the channel again
    channel_identifier4 = create_settled_channel(
        A,
        pending_transfers_tree_2.locked_amount,
        pending_transfers_tree_2.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    # Both old and new unlocks should go through
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier4, B, A, pending_transfers_tree_2.packed_transfers
        ),
        {"from": A},
    )
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier3, B, A, pending_transfers_tree_1.packed_transfers
        ),
        {"from": A},
    )


def test_channel_unlock_registered_expired_lock_refunds(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    secret_registry_contract: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    close_and_update_channel: Callable,
) -> None:
    """unlock() should refund tokens locked with secrets revealed after the expiration"""
    (A, B) = get_accounts(2)
    max_lock_expiration = 3
    settle_timeout = 8

    values_A = ChannelValues(deposit=20, transferred=5)
    values_B = ChannelValues(deposit=30, transferred=40)

    # Create channel and deposit
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, values_A.deposit, B)
    channel_deposit(channel_identifier, B, values_B.deposit, A)

    # Mock pending transfers data
    pending_transfers_tree = get_pending_transfers_tree(
        web3,
        [1, 3, 5],
        [2, 4],
        min_expiration_delta=max_lock_expiration - 2,
        max_expiration_delta=max_lock_expiration,
    )
    values_B.locksroot = pending_transfers_tree.hash_of_packed_transfers
    values_B.locked_amounts = LockedAmounts(
        claimable_locked=get_locked_amount(pending_transfers_tree.transfers)
    )

    # Locks expire
    mine_blocks(web3, max_lock_expiration)

    # Secrets are revealed before settlement window, but after expiration
    for (_, _, secrethash, secret) in pending_transfers_tree.unlockable:
        call_and_transact(secret_registry_contract.functions.registerSecret(secret), {"from": A})
        assert (
            secret_registry_contract.functions.getSecretRevealBlockHeight(secrethash).call()
            == web3.eth.block_number
        )

    close_and_update_channel(channel_identifier, A, values_A, B, values_B)
    mine_blocks(web3, settle_timeout)

    # settle channel
    call_settle(token_network, channel_identifier, A, values_A, B, values_B)

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # Unlock works after channel is settled
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        )
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # check that all tokens have been refunded, as locks have expired already
    assert balance_A == pre_balance_A
    assert balance_B == pre_balance_B + values_B.locked_amounts.locked
    assert balance_contract == pre_balance_contract - values_B.locked_amounts.locked


def test_channel_unlock_unregistered_locks(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    close_and_update_channel: Callable,
    custom_token: Contract,
) -> None:
    """unlock() should refund tokens locked by secrets not registered before settlement"""
    (A, B) = get_accounts(2)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN

    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    locked_A = pending_transfers_tree.locked_amount
    (vals_A, vals_B) = (
        ChannelValues(
            deposit=35,
            withdrawn=10,
            transferred=0,
            locked_amounts=LockedAmounts(claimable_locked=locked_A),
        ),
        ChannelValues(deposit=40, withdrawn=10, transferred=20),
    )

    vals_A.locksroot = pending_transfers_tree.hash_of_packed_transfers
    vals_B.locksroot = fake_bytes(32, "03")
    channel_identifier = create_channel_and_deposit(A, B, vals_A.deposit, vals_B.deposit)
    withdraw_channel(channel_identifier, A, vals_A.withdrawn, UINT256_MAX, B)
    withdraw_channel(channel_identifier, B, vals_B.withdrawn, UINT256_MAX, A)

    close_and_update_channel(channel_identifier, A, vals_A, B, vals_B)

    # Secret hasn't been registered before settlement timeout
    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    # Someone unlocks A's pending transfers - all tokens should be refunded
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree.packed_transfers
        ),
        {"from": A},
    )

    # A gets back locked tokens
    assert (
        custom_token.functions.balanceOf(A).call()
        == vals_A.deposit - vals_A.transferred + vals_B.transferred
    )


def test_channel_unlock_before_settlement_fails(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    close_and_update_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """unlock() should not work before settlement"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    values_A = ChannelValues(deposit=20, transferred=5)
    values_B = ChannelValues(deposit=30, transferred=40)

    # Create channel and deposit
    channel_identifier = create_channel(A, B, settle_timeout)[0]

    # Mock pending transfers data
    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    values_B.locksroot = pending_transfers_tree.hash_of_packed_transfers
    values_B.locked_amounts = LockedAmounts(
        claimable_locked=get_locked_amount(pending_transfers_tree.transfers)
    )

    # Reveal secrets before settlement window ends
    reveal_secrets(A, pending_transfers_tree.unlockable)

    # Unlock fails before channel is not settled
    with pytest.raises(TransactionFailed, match="TN/unlock: channel id still exists"):
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        ).call()

    channel_deposit(channel_identifier, A, values_A.deposit, B)
    channel_deposit(channel_identifier, B, values_B.deposit, A)

    # Unlock fails before channel is not settled
    with pytest.raises(TransactionFailed, match="TN/unlock: channel id still exists"):
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        ).call()

    close_and_update_channel(channel_identifier, A, values_A, B, values_B)

    # Unlock fails before settlement window is over and channel is not settled
    with pytest.raises(TransactionFailed, match="TN/unlock: channel id still exists"):
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        ).call()

    # Settlement window must be over before settling the channel
    mine_blocks(web3, settle_timeout)

    # Unlock fails before settle is called
    with pytest.raises(TransactionFailed, match="TN/unlock: channel id still exists"):
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        ).call()

    # settle channel
    call_settle(token_network, channel_identifier, A, values_A, B, values_B)

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # Unlock works after channel is settled
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        )
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_A == pre_balance_A + 9
    assert balance_B == pre_balance_B + 6
    assert balance_contract == pre_balance_contract - values_B.locked_amounts.locked


def test_unlock_fails_with_partial_locks(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_settled_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """unlock() should fail when one lock is missing"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree.unlockable)

    # Settle the channel
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree.locked_amount,
        pending_transfers_tree.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    # Unlock with one leave missing does not work
    types = ["uint256", "uint256", "bytes32"]
    for index in range(len(pending_transfers_tree.transfers)):
        pending_transfers = list(pending_transfers_tree.transfers)
        del pending_transfers[index]
        packed_transfers_tampered = get_packed_transfers(tuple(pending_transfers), types)
        with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
            token_network.functions.unlock(
                channel_identifier, B, A, packed_transfers_tampered
            ).call({"from": A})

    # Unlock with full pending locks does work
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree.packed_transfers
        ),
        {"from": A},
    )


def test_unlock_tampered_proof_fails(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_settled_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """unlock() should fail when the submitted proofs are tampered"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree.unlockable)

    # Settle the channel
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree.locked_amount,
        pending_transfers_tree.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    # Unlock with tampered locks does not work
    types = ["uint256", "uint256", "bytes32"]
    for index in range(len(pending_transfers_tree.transfers)):
        pending_transfers = list(pending_transfers_tree.transfers)
        pending_transfers[index][2:] = random_secret()
        packed_transfers_tampered = get_packed_transfers(tuple(pending_transfers), types)
        with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
            token_network.functions.unlock(
                channel_identifier, B, A, packed_transfers_tampered
            ).call({"from": A})

    # Unlock with correct pending locks does work
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree.packed_transfers
        ),
        {"from": A},
    )


def test_channel_unlock_both_participants(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    secret_registry_contract: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    close_and_update_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """A scenario where both parties get some of the pending transfers"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    values_A = ChannelValues(deposit=100, transferred=5)
    values_B = ChannelValues(deposit=100, transferred=40)

    # Create channel and deposit
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, values_A.deposit, B)
    channel_deposit(channel_identifier, B, values_B.deposit, A)

    # Mock pending transfers data for A
    pending_transfers_tree_A = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    values_A.locksroot = pending_transfers_tree_A.hash_of_packed_transfers
    values_A.locked_amounts = LockedAmounts(
        claimable_locked=get_locked_amount(pending_transfers_tree_A.transfers)
    )

    # Reveal A's secrets before settlement window ends
    reveal_secrets(A, pending_transfers_tree_A.unlockable)

    # Mock pending transfers data for B
    pending_transfers_tree_B = get_pending_transfers_tree(web3, [2, 4, 6], [5, 10], settle_timeout)
    values_B.locksroot = pending_transfers_tree_B.hash_of_packed_transfers
    values_B.locked_amounts = LockedAmounts(
        claimable_locked=get_locked_amount(pending_transfers_tree_B.transfers)
    )

    # Reveal B's secrets before settlement window ends
    reveal_secrets(B, pending_transfers_tree_B.unlockable)

    close_and_update_channel(channel_identifier, A, values_A, B, values_B)

    # Settle channel
    mine_blocks(web3, settle_timeout)

    call_settle(token_network, channel_identifier, A, values_A, B, values_B)

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # A unlock's
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree_B.packed_transfers
        )
    )

    # B unlock's
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        )
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # Unlocked pending transfers A -> B, that belong to B
    unlockable_A = get_unlocked_amount(
        secret_registry_contract, pending_transfers_tree_A.packed_transfers
    )

    # Expired pending transfers A -> B, that belong to A
    expired_A = get_locked_amount(pending_transfers_tree_A.expired)

    # Unlocked pending transfers B -> A, that belong to A
    unlockable_B = get_unlocked_amount(
        secret_registry_contract, pending_transfers_tree_B.packed_transfers
    )

    # Expired pending transfers B -> A, that belong to B
    expired_B = get_locked_amount(pending_transfers_tree_B.expired)

    # check that A and B both received the expected amounts
    assert balance_contract == (
        pre_balance_contract - values_B.locked_amounts.locked - values_A.locked_amounts.locked
    )
    assert balance_A == pre_balance_A + unlockable_B + expired_A
    assert balance_B == pre_balance_B + unlockable_A + expired_B


def test_unlock_twice_fails(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_settled_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """The same unlock() call twice do not work"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree_1 = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_1.unlockable)

    # Settle the channel
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree_1.locked_amount,
        pending_transfers_tree_1.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_1.packed_transfers
        ),
        {"from": A},
    )

    # Calling unlock twice does not work
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_1.packed_transfers
        ).call({"from": A})


def test_unlock_no_locks(
    web3: Web3, token_network: Contract, get_accounts: Callable, create_settled_channel: Callable
) -> None:
    """unlock() should work on no pending locks"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree = get_pending_transfers_tree(
        web3=web3,
        unlockable_amounts=[],
        expired_amounts=[2, 4],
        min_expiration_delta=settle_timeout,
    )

    # Settle the channel
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree.locked_amount,
        pending_transfers_tree.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree.packed_transfers
        ),
        {"from": A},
    )

    # Calling unlock twice does not work
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree.packed_transfers
        ).call({"from": A})


def test_channel_unlock_with_a_large_expiration(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    close_and_update_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """unlock() should still work after a delayed settleChannel() call"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    values_A = ChannelValues(deposit=20, transferred=5)
    values_B = ChannelValues(deposit=30, transferred=40)

    # Create channel and deposit
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, values_A.deposit, B)
    channel_deposit(channel_identifier, B, values_B.deposit, A)

    # Mock pending transfers data with large expiration date
    pending_transfers_tree = get_pending_transfers_tree(
        web3, [1, 3, 5], [2, 4], settle_timeout + 100
    )
    values_B.locksroot = pending_transfers_tree.hash_of_packed_transfers
    values_B.locked_amounts = LockedAmounts(
        claimable_locked=get_locked_amount(pending_transfers_tree.transfers)
    )

    # Reveal secrets before settlement window ends
    reveal_secrets(A, pending_transfers_tree.unlockable)

    close_and_update_channel(channel_identifier, A, values_A, B, values_B)

    # Settle channel after a "long" time
    mine_blocks(web3, settle_timeout + 50)

    call_settle(token_network, channel_identifier, A, values_A, B, values_B)

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # Unlock the tokens must still work
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        )
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()
    assert balance_A == pre_balance_A + 9
    assert balance_B == pre_balance_B + 6
    assert balance_contract == pre_balance_contract - values_B.locked_amounts.locked


def test_reverse_participants_unlock(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_settled_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """unlock() with wrong argument orders"""
    (A, B, C) = get_accounts(3)
    settle_timeout = 12

    # Mock pending transfers data
    pending_transfers_tree_A = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    pending_transfers_tree_B = get_pending_transfers_tree(web3, [3, 4, 6], [4], settle_timeout)
    assert (
        pending_transfers_tree_A.hash_of_packed_transfers
        != pending_transfers_tree_B.hash_of_packed_transfers
    )

    # Reveal secrets
    reveal_secrets(A, pending_transfers_tree_A.unlockable)
    reveal_secrets(B, pending_transfers_tree_B.unlockable)

    # Settle the channel
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree_A.locked_amount,
        pending_transfers_tree_A.hash_of_packed_transfers,
        B,
        pending_transfers_tree_B.locked_amount,
        pending_transfers_tree_B.hash_of_packed_transfers,
        settle_timeout,
    )

    # A trying to unlock its own locksroot & locked amount MUST fail
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree_A.packed_transfers
        ).call({"from": A})

    # Delegate trying to unlock A's own locksroot & locked amount on behalf of A MUST fail
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree_A.packed_transfers
        ).call({"from": C})

    # B trying to unlock its own locksroot & locked amount MUST fail
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_B.packed_transfers
        ).call({"from": B})

    # Delegate trying to unlock B's own locksroot & locked amount on behalf of B MUST fail
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_B.packed_transfers
        ).call({"from": B})

    with pytest.raises(TransactionFailed, match="TN: identical addresses"):
        token_network.functions.unlock(
            channel_identifier, A, A, pending_transfers_tree_A.packed_transfers
        ).call({"from": A})

    with pytest.raises(TransactionFailed, match="TN: identical addresses"):
        token_network.functions.unlock(
            channel_identifier, B, B, pending_transfers_tree_B.packed_transfers
        ).call({"from": B})

    # Someone trying to unlock B's locksroot & locked amount on behalf of A MUST succeed
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree_B.packed_transfers
        ),
        {"from": C},
    )

    # Someone trying to unlock A's locksroot & locked amount on behalf of B MUST succeed
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        ),
        {"from": C},
    )


def test_unlock_different_channel_same_participants_fail(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_settled_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """Try to confuse unlock() with two channels between the same participants"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    # Mock pending transfers data
    pending_transfers_tree_1 = get_pending_transfers_tree(web3, [1, 3, 5], [2, 4], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_1.unlockable)
    channel_identifier = create_settled_channel(
        A,
        pending_transfers_tree_1.locked_amount,
        pending_transfers_tree_1.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    # The first channel is settled, so we create another one
    pending_transfers_tree_2 = get_pending_transfers_tree(web3, [3, 5], [2, 4, 3], settle_timeout)
    reveal_secrets(A, pending_transfers_tree_2.unlockable)
    channel_identifier2 = create_settled_channel(
        A,
        pending_transfers_tree_2.locked_amount,
        pending_transfers_tree_2.hash_of_packed_transfers,
        B,
        0,
        LOCKSROOT_OF_NO_LOCKS,
        settle_timeout,
    )

    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_2.packed_transfers
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/unlock: locksroot mismatch"):
        token_network.functions.unlock(
            channel_identifier2, B, A, pending_transfers_tree_1.packed_transfers
        ).call({"from": A})

    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_1.packed_transfers
        ),
        {"from": A},
    )
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier2, B, A, pending_transfers_tree_2.packed_transfers
        ),
        {"from": A},
    )


def test_unlock_channel_event(
    web3: Web3,
    token_network: Contract,
    secret_registry_contract: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    close_and_update_channel: Callable,
    reveal_secrets: Callable,
    event_handler: Callable,
) -> None:
    """Successful unlock() should cause an UNLOCKED event"""
    (A, B) = get_accounts(2)
    settle_timeout = 8

    values_A = ChannelValues(deposit=20, transferred=5)
    values_B = ChannelValues(deposit=30, transferred=40)

    # Create channel and deposit
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, values_A.deposit, B)
    channel_deposit(channel_identifier, B, values_B.deposit, A)

    # Mock pending transfers data
    pending_transfers_tree = get_pending_transfers_tree(
        web3, [1, 3, 5], [2, 4], settle_timeout + 100
    )
    values_B.locksroot = pending_transfers_tree.hash_of_packed_transfers
    values_B.locked_amounts = LockedAmounts(
        claimable_locked=get_locked_amount(pending_transfers_tree.transfers)
    )

    # Reveal secrets before settlement window ends
    reveal_secrets(A, pending_transfers_tree.unlockable)

    close_and_update_channel(channel_identifier, A, values_A, B, values_B)

    # Settlement window must be over before settling the channel
    mine_blocks(web3, settle_timeout)

    call_settle(token_network, channel_identifier, A, values_A, B, values_B)

    ev_handler = event_handler(token_network)

    # Unlock the tokens
    txn_hash = call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree.packed_transfers
        )
    )

    unlocked_amount = get_unlocked_amount(
        secret_registry_contract, pending_transfers_tree.packed_transfers
    )

    # Add event
    ev_handler.add(
        txn_hash,
        ChannelEvent.UNLOCKED,
        check_channel_unlocked(
            channel_identifier,
            A,
            B,
            values_B.locksroot,
            unlocked_amount,
            values_B.locked_amounts.locked - unlocked_amount,
        ),
    )

    # Check that event was properly emitted
    ev_handler.check()



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_CUSTOM_TOKEN, CONTRACT_HUMAN_STANDARD_TOKEN
from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.utils import call_and_transact


def test_token_mint(web3: Web3, custom_token: Contract, get_accounts: Callable) -> None:
    """Use the mint() function of the custom token contract"""

    (A, B) = get_accounts(2)
    token = custom_token
    multiplier = custom_token.functions.multiplier().call()
    supply = token.functions.totalSupply().call()

    token_pre_balance = web3.eth.get_balance(token.address)
    tokens_a = 50 * multiplier
    call_and_transact(token.functions.mint(tokens_a), {"from": A})
    assert token.functions.balanceOf(A).call() == tokens_a
    assert token.functions.balanceOf(B).call() == 0
    assert token.functions.totalSupply().call() == supply + tokens_a
    assert web3.eth.get_balance(token.address) == token_pre_balance

    tokens_b = 50 * multiplier
    call_and_transact(token.functions.mintFor(tokens_b, B), {"from": A})
    assert token.functions.balanceOf(A).call() == tokens_a
    assert token.functions.balanceOf(B).call() == tokens_b
    assert token.functions.totalSupply().call() == supply + tokens_a + tokens_b
    assert web3.eth.get_balance(token.address) == token_pre_balance


def test_approve_transfer(custom_token: Contract, get_accounts: Callable) -> None:
    """Use the approve() function of the custom token contract"""

    (A, B) = get_accounts(2)
    token = custom_token
    call_and_transact(token.functions.mint(50), {"from": A})
    initial_balance_A = token.functions.balanceOf(A).call()
    initial_balance_B = token.functions.balanceOf(B).call()
    to_transfer = 20
    call_and_transact(token.functions.approve(B, to_transfer), {"from": A})
    call_and_transact(token.functions.transferFrom(A, B, to_transfer), {"from": B})
    assert token.functions.balanceOf(B).call() == initial_balance_B + to_transfer
    assert token.functions.balanceOf(A).call() == initial_balance_A - to_transfer

    assert custom_token.functions.allowance(_owner=A, _spender=B).call() == 0
    assert call_and_transact(custom_token.functions.approve(_spender=B, _value=25), {"from": A})
    assert custom_token.functions.allowance(_owner=A, _spender=B).call() == 25
    assert custom_token.functions.allowance(_owner=A, _spender=token.address).call() == 0


def test_token_transfer_funds(web3: Web3, custom_token: Contract, get_accounts: Callable) -> None:
    """transferFunds() should fail when the ETH balance of the contract is zero"""

    A = get_accounts(1)[0]
    token = custom_token
    multiplier = custom_token.functions.multiplier().call()
    assert multiplier > 0
    supply = token.functions.totalSupply().call()
    assert supply > 0

    owner = custom_token.functions.owner_address().call()

    assert web3.eth.get_balance(token.address) == 0
    with pytest.raises(TransactionFailed):
        token.functions.transferFunds().call({"from": owner})

    call_and_transact(token.functions.mint(50), {"from": A})
    assert web3.eth.get_balance(token.address) == 0


def test_custom_token(
    custom_token: Contract, web3: Web3, contracts_manager: ContractManager
) -> None:
    """See custom_token.address contains the expected code"""
    blockchain_bytecode = web3.eth.get_code(custom_token.address)
    compiled_bytecode = contracts_manager.get_runtime_hexcode(CONTRACT_CUSTOM_TOKEN)
    assert blockchain_bytecode.hex() == compiled_bytecode


def test_human_standard_token(
    human_standard_token: Contract, web3: Web3, contracts_manager: ContractManager
) -> None:
    """See human_standard_token.address contains the expected code"""
    blockchain_bytecode = web3.eth.get_code(human_standard_token.address)
    compiled_bytecode = contracts_manager.get_runtime_hexcode(CONTRACT_HUMAN_STANDARD_TOKEN)
    assert blockchain_bytecode.hex() == compiled_bytecode



import logging
from typing import Generator

import pytest
from eth_tester import EthereumTester, PyEVMBackend
from web3 import Web3
from web3.providers.eth_tester import EthereumTesterProvider

from raiden_contracts.tests.utils.constants import (
    FAUCET_ADDRESS,
    FAUCET_ALLOWANCE,
    FAUCET_PRIVATE_KEY,
)

log = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def ethereum_tester(patch_genesis_gas_limit: None) -> EthereumTester:  # pylint: disable=W0613
    """Returns an instance of an Ethereum tester"""
    return EthereumTester(PyEVMBackend())


@pytest.fixture(scope="session")
def patch_genesis_gas_limit() -> None:
    """Increases the block gas limit, to make the TokenNetworkRegistry contract deployable"""

    tmp_limit = 6 * 10 ** 6
    import eth_tester.backends.pyevm.main as pyevm_main

    pyevm_main.GENESIS_GAS_LIMIT = tmp_limit
    import eth.vm.forks.frontier.headers as headers

    headers.GENESIS_GAS_LIMIT = tmp_limit


@pytest.fixture(scope="session")
def web3(ethereum_tester: EthereumTester) -> Web3:
    """Returns an initialized Web3 instance"""
    provider = EthereumTesterProvider(ethereum_tester)
    web3 = Web3(provider)
    # Improve test speed by skipping the gas cost estimation.
    web3.eth.estimateGas = lambda txn: int(5.2e6)  # pylint: disable=E1101

    # add faucet account to tester
    ethereum_tester.add_account(FAUCET_PRIVATE_KEY.hex())

    # make faucet rich
    ethereum_tester.send_transaction(
        {
            "from": ethereum_tester.get_accounts()[0],
            "to": FAUCET_ADDRESS,
            "gas": 21000,
            "value": FAUCET_ALLOWANCE,
        }
    )

    # Enable strict type checks
    web3.enable_strict_bytes_type_checking()

    return web3


@pytest.fixture(autouse=True)
def auto_revert_chain(web3: Web3) -> Generator:
    """Reverts the chain to its before the test run

    This reverts the side effects created during the test run, so that we can
    reuse the same chain and contract deployments for other tests.

    This is useful especially when using ethereum tester - its log filtering
    is very slow once enough events are present on-chain.
    """
    snapshot_id = web3.testing.snapshot()  # type: ignore
    yield
    web3.testing.revert(snapshot_id)  # type: ignore



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3.contract import Contract
from web3.exceptions import ValidationError

from raiden_contracts.constants import EMPTY_ADDRESS, ChannelEvent
from raiden_contracts.tests.utils import EMPTY_LOCKSROOT, call_and_transact
from raiden_contracts.utils.events import check_channel_settled, check_withdraw_2
from raiden_contracts.utils.type_aliases import BlockExpiration

EXPIRATION = BlockExpiration(100)


def test_cooperative_settle_channel_call(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 5
    balance_B = 25

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B, EXPIRATION
    )

    # negative balances
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, -1, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, -1, EXPIRATION, signature_B2, signature_A2),
        )
    # wrong participant addresses
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (0x0, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (0x0, balance_B, EXPIRATION, signature_B2, signature_A2),
        )
    # invalid signatures
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, 0x0, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, 0x0),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, 0x0, signature_A2),
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, 0x0),
        )
    # empty addresses
    with pytest.raises(TransactionFailed, match="TN: participant address zero"):
        call_and_transact(
            token_network.functions.cooperativeSettle(
                channel_identifier,
                (EMPTY_ADDRESS, balance_A, EXPIRATION, signature_A1, signature_B1),
                (B, balance_B, EXPIRATION, signature_B2, signature_A2),
            ),
            {"from": C},
        )
    with pytest.raises(TransactionFailed, match="TN: partner address zero"):
        call_and_transact(
            token_network.functions.cooperativeSettle(
                channel_identifier,
                (A, balance_A, EXPIRATION, signature_A1, signature_B1),
                (EMPTY_ADDRESS, balance_B, EXPIRATION, signature_B2, signature_A2),
            ),
            {"from": C},
        )

    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ),
        {"from": C},
    )


def test_cooperative_settle_channel_signatures(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 4
    balance_B = 26

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A1, signature_B1, signature_C1) = create_withdraw_signatures(
        [A, B, C], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2, signature_C2) = create_withdraw_signatures(
        [A, B, C], channel_identifier, B, balance_B, EXPIRATION
    )

    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_C1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ).call({"from": C})
    with pytest.raises(TransactionFailed, match="TN/withdraw: channel id mismatch"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_C1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ).call({"from": C})
    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_C2, signature_A2),
        ).call({"from": C})
    with pytest.raises(TransactionFailed, match="TN/withdraw: channel id mismatch"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_C2),
        ).call({"from": C})

    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_B, EXPIRATION, signature_A1, signature_B1),
            (B, balance_A, EXPIRATION, signature_B2, signature_A2),
        ).call({"from": C})

    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ),
        {"from": C},
    )


def test_cooperative_settle_channel_0(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 0
    balance_B = 30

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B, EXPIRATION
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ),
        {"from": C},
    )

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


def test_cooperative_settle_channel_00(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 0
    deposit_B = 0
    balance_A = 0
    balance_B = 0

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B, EXPIRATION
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ),
        {"from": C},
    )

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


def test_cooperative_settle_channel_state(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 5
    balance_B = 25
    assert deposit_A + deposit_B == balance_A + balance_B

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B, EXPIRATION
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ),
        {"from": C},
    )

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


def test_cooperative_settle_channel_state_withdraw(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    withdraw_A = 3
    withdraw_B = 7
    balance_A = 5
    balance_B = 15
    assert deposit_A + deposit_B == withdraw_A + withdraw_B + balance_A + balance_B

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    withdraw_channel(channel_identifier, A, withdraw_A, EXPIRATION, B)
    withdraw_channel(channel_identifier, B, withdraw_B, EXPIRATION, A)

    # We need to add the already withdrawn amount to the balance as `withdraw_amount`
    # is a monotonic value
    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A + withdraw_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B + withdraw_B, EXPIRATION
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A + withdraw_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B + withdraw_B, EXPIRATION, signature_B2, signature_A2),
        ),
        {"from": C},
    )

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


def test_cooperative_settle_channel_smaller_total_amount(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 15 - 1
    balance_B = 15

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B, EXPIRATION
    )

    with pytest.raises(TransactionFailed, match="TN/coopSettle: incomplete amount"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ).call({"from": C})


def test_cooperative_settle_channel_bigger_withdraw(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    withdraw_A = 3
    withdraw_B = 7
    balance_A = 6
    balance_B = 15
    assert deposit_A + deposit_B < withdraw_A + withdraw_B + balance_A + balance_B

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    withdraw_channel(channel_identifier, A, withdraw_A, EXPIRATION, B)
    withdraw_channel(channel_identifier, B, withdraw_B, EXPIRATION, A)

    # We need to add the already withdrawn amount to the balance as `withdraw_amount`
    # is a monotonic value
    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A + withdraw_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B + withdraw_B, EXPIRATION
    )

    with pytest.raises(TransactionFailed, match="TN/coopSettle: incomplete amount"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A + withdraw_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B + withdraw_B, EXPIRATION, signature_B2, signature_A2),
        ).call({"from": C})


def test_cooperative_settle_channel_wrong_balances(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 7
    balance_B = 23

    balance_A_fail1 = 20
    balance_B_fail1 = 11
    balance_A_fail2 = 6
    balance_B_fail2 = 8

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A1_fail1, signature_B1_fail1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A_fail1, EXPIRATION
    )
    (signature_A2_fail1, signature_B2_fail1) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B_fail1, EXPIRATION
    )

    with pytest.raises(TransactionFailed, match="TN/coopSettle: incomplete amount"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A_fail1, EXPIRATION, signature_A1_fail1, signature_B1_fail1),
            (B, balance_B_fail1, EXPIRATION, signature_B2_fail1, signature_A2_fail1),
        ).call({"from": C})

    (signature_A1_fail2, signature_B1_fail2) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A_fail2, EXPIRATION
    )
    (signature_A2_fail2, signature_B2_fail2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B_fail2, EXPIRATION
    )

    with pytest.raises(TransactionFailed, match="TN/coopSettle: incomplete amount"):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A_fail1, EXPIRATION, signature_A1_fail2, signature_B1_fail2),
            (B, balance_B_fail1, EXPIRATION, signature_B2_fail2, signature_A2_fail2),
        ).call({"from": C})

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B, EXPIRATION
    )
    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
        ),
        {"from": C},
    )


def test_cooperative_close_replay_reopened_channel(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    (A, B) = get_accounts(2)
    deposit_A = 15
    deposit_B = 10
    balance_A = 2
    balance_B = 23

    channel_identifier1 = create_channel(A, B)[0]
    channel_deposit(channel_identifier1, A, deposit_A, B)
    channel_deposit(channel_identifier1, B, deposit_B, A)

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier1, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier1, B, balance_B, EXPIRATION
    )

    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier1,
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
        ),
        {"from": B},
    )

    # Reopen the channel and make sure we cannot use the old balance proof
    channel_identifier2 = create_channel(A, B)[0]
    channel_deposit(channel_identifier2, A, deposit_A, B)
    channel_deposit(channel_identifier2, B, deposit_B, A)

    assert channel_identifier1 != channel_identifier2
    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.cooperativeSettle(
            channel_identifier2,
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
        ).call({"from": B})

    # Signed message with the correct channel identifier must work
    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier2, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier2, B, balance_B, EXPIRATION
    )
    call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier2,
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
        ),
        {"from": B},
    )


def test_cooperative_settle_channel_event(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_withdraw_signatures: Callable,
    event_handler: Callable,
) -> None:
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)
    deposit_A = 10
    balance_A = 2
    balance_B = 8

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)

    (signature_A1, signature_B1) = create_withdraw_signatures(
        [A, B], channel_identifier, A, balance_A, EXPIRATION
    )
    (signature_A2, signature_B2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, balance_B, EXPIRATION
    )

    txn_hash = call_and_transact(
        token_network.functions.cooperativeSettle(
            channel_identifier,
            (B, balance_B, EXPIRATION, signature_B2, signature_A2),
            (A, balance_A, EXPIRATION, signature_A1, signature_B1),
        ),
        {"from": B},
    )

    withdraw_events = token_network.events.ChannelWithdraw.getLogs()
    assert any(map(check_withdraw_2(channel_identifier, A, balance_A), withdraw_events))
    assert any(map(check_withdraw_2(channel_identifier, B, balance_B), withdraw_events))

    ev_handler.add(
        txn_hash,
        ChannelEvent.SETTLED,
        check_channel_settled(
            channel_identifier,
            B,
            balance_B,
            EMPTY_LOCKSROOT,
            A,
            balance_A,
            EMPTY_LOCKSROOT,
        ),
    )
    ev_handler.check()



from typing import Callable, List

import pytest
from eth_tester.exceptions import TransactionFailed
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import OneToNEvent
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.utils.proofs import sign_one_to_n_iou
from raiden_contracts.utils.type_aliases import BlockExpiration, ChainID, TokenAmount


def test_claim(
    user_deposit_contract: Contract,
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    web3: Web3,
    event_handler: Callable,
    create_account: Callable,
    create_service_account: Callable,
    make_iou: Callable,
) -> None:
    ev_handler = event_handler(one_to_n_contract)
    A = create_account()
    B = create_service_account()
    deposit_to_udc(A, 30)

    # IOU expired
    with pytest.raises(TransactionFailed, match="IOU expired"):
        bad_expiration = web3.eth.block_number - 1
        one_to_n_contract.functions.claim(
            **make_iou(sender=A, receiver=B, expiration_block=bad_expiration)
        ).call({"from": A})

    # Wrong OneToN address
    with pytest.raises(TransactionFailed, match="Signature mismatch"):
        one_to_n_contract.functions.claim(
            **make_iou(sender=A, receiver=B, one_to_n_address=A)
        ).call({"from": A})

    # Wrong chain_id
    with pytest.raises(TransactionFailed, match="Signature mismatch"):
        one_to_n_contract.functions.claim(**make_iou(sender=A, receiver=B, chain_id=77)).call(
            {"from": A}
        )

    # bad signature
    with pytest.raises(TransactionFailed, match="Signature mismatch"):
        iou = make_iou(sender=A, receiver=B, amount=10)
        iou2 = make_iou(sender=A, receiver=B, amount=11)
        iou["signature"] = iou2["signature"]  # use signature for wrong amount
        one_to_n_contract.functions.claim(**iou).call({"from": A})

    # happy case
    iou = make_iou(sender=A, receiver=B)
    tx_hash = call_and_transact(one_to_n_contract.functions.claim(**iou), {"from": A})

    ev_handler.assert_event(
        tx_hash,
        OneToNEvent.CLAIMED,
        dict(
            sender=A,
            receiver=B,
            expiration_block=iou["expiration_block"],
            transferred=iou["amount"],
        ),
    )
    assert user_deposit_contract.functions.balances(A).call() == 20
    assert user_deposit_contract.functions.balances(B).call() == 10

    # can't be claimed twice
    with pytest.raises(TransactionFailed, match="Already settled session"):
        one_to_n_contract.functions.claim(**iou).call({"from": A})


def test_bulk_claim_happy_path(
    user_deposit_contract: Contract,
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    make_iou: Callable,
    create_service_account: Callable,
) -> None:
    (A, B) = get_accounts(2)
    deposit_to_udc(A, 30)
    deposit_to_udc(B, 30)
    C = create_service_account()

    def bulk_claim(ious: List[dict]) -> HexBytes:
        return call_and_transact(
            one_to_n_contract.functions.bulkClaim(
                senders=[x["sender"] for x in ious],
                receivers=[x["receiver"] for x in ious],
                amounts=[x["amount"] for x in ious],
                expiration_blocks=[x["expiration_block"] for x in ious],
                signatures=b"".join(x["signature"] for x in ious),
            ),
            {"from": A},
        )

    ious = [make_iou(A, C, amount=10), make_iou(B, C, amount=20)]
    bulk_claim(ious)

    assert user_deposit_contract.functions.balances(A).call() == 20
    assert user_deposit_contract.functions.balances(B).call() == 10
    assert user_deposit_contract.functions.balances(C).call() == 30


def test_bulk_claim_errors(
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    create_service_account: Callable,
    make_iou: Callable,
) -> None:
    (A, B) = get_accounts(2)
    C = create_service_account()
    deposit_to_udc(A, 30)
    deposit_to_udc(B, 30)

    ious = [make_iou(A, C), make_iou(B, C)]
    senders = [x["sender"] for x in ious]
    receivers = [x["receiver"] for x in ious]
    amounts = [x["amount"] for x in ious]
    expiration_blocks = [x["expiration_block"] for x in ious]
    signatures = b"".join(x["signature"] for x in ious)

    # One value too many to `amounts`
    with pytest.raises(
        TransactionFailed,
        match="Same number of elements required for all input parameters",
    ):
        call_and_transact(
            one_to_n_contract.functions.bulkClaim(
                senders=senders,
                receivers=receivers,
                amounts=amounts + [1],
                expiration_blocks=expiration_blocks,
                signatures=signatures,
            ),
            {"from": A},
        )

    # One byte too few/many in `signatures`
    for sig in [signatures + b"1", signatures[:-1]]:
        with pytest.raises(
            TransactionFailed, match="`signatures` should contain 65 bytes per IOU"
        ):
            call_and_transact(
                one_to_n_contract.functions.bulkClaim(
                    senders=senders,
                    receivers=receivers,
                    amounts=amounts,
                    expiration_blocks=expiration_blocks,
                    signatures=sig,
                ),
                {"from": A},
            )

    # Cause a signature mismatch by changing one amount
    with pytest.raises(TransactionFailed, match="Signature mismatch"):
        call_and_transact(
            one_to_n_contract.functions.bulkClaim(
                senders=senders,
                receivers=receivers,
                amounts=[amounts[0], amounts[1] + 1],
                expiration_blocks=expiration_blocks,
                signatures=signatures,
            ),
            {"from": A},
        )


def test_getSingleSignature(one_to_n_internals: Contract) -> None:
    signatures = bytes(range(65 * 3))
    assert (
        one_to_n_internals.functions.getSingleSignaturePublic(signatures, 0).call()
        == signatures[:65]
    )
    assert (
        one_to_n_internals.functions.getSingleSignaturePublic(signatures, 1).call()
        == signatures[65 : 65 * 2]
    )
    assert (
        one_to_n_internals.functions.getSingleSignaturePublic(signatures, 2).call()
        == signatures[65 * 2 : 65 * 3]
    )


def test_claim_by_unregistered_service(
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    get_private_key: Callable,
    web3: Web3,
) -> None:
    """OneToN contract should not work for an unregistered service provider."""
    (A, B) = get_accounts(2)
    deposit_to_udc(A, 30)

    amount = TokenAmount(10)
    expiration = BlockExpiration(web3.eth.block_number + 2)
    chain_id = web3.eth.chain_id

    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=ChainID(chain_id),
    )

    # Doesn't work because B is not registered
    with pytest.raises(TransactionFailed, match="receiver not registered"):
        call_and_transact(
            one_to_n_contract.functions.claim(
                sender=A,
                receiver=B,
                amount=amount,
                expiration_block=expiration,
                signature=signature,
            ),
            {"from": A},
        )


def test_claim_with_insufficient_deposit(
    user_deposit_contract: Contract,
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    get_private_key: Callable,
    web3: Web3,
    event_handler: Callable,
    create_account: Callable,
    create_service_account: Callable,
) -> None:
    ev_handler = event_handler(one_to_n_contract)
    A = create_account()
    B = create_service_account()
    deposit_to_udc(A, 6)
    chain_id = web3.eth.chain_id

    amount = TokenAmount(10)
    expiration = BlockExpiration(web3.eth.block_number + 1)
    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=ChainID(chain_id),
    )

    # amount is 10, but only 6 are in deposit
    # check return value (transactions don't give back return values, so use call)
    assert (
        one_to_n_contract.functions.claim(A, B, amount, expiration, signature).call({"from": A})
        == 6
    )
    # check that transaction succeeds
    call_and_transact(
        one_to_n_contract.functions.claim(A, B, amount, expiration, signature),
        {"from": A},
    )

    assert user_deposit_contract.functions.balances(A).call() == 0
    assert user_deposit_contract.functions.balances(B).call() == 6

    # claim can be retried when transferred amount was 0
    expiration = BlockExpiration(web3.eth.block_number + 10)
    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=ChainID(chain_id),
    )
    call_and_transact(
        one_to_n_contract.functions.claim(A, B, amount, expiration, signature),
        {"from": A},
    )
    deposit_to_udc(A, 6 + 4)
    tx_hash = call_and_transact(
        one_to_n_contract.functions.claim(A, B, amount, expiration, signature),
        {"from": A},
    )
    ev_handler.assert_event(
        tx_hash,
        OneToNEvent.CLAIMED,
        dict(sender=A, receiver=B, expiration_block=expiration, transferred=4),
    )



import json
from sys import argv
from tempfile import NamedTemporaryFile
from typing import IO, Callable, Dict, Iterable, List, Optional

import pytest
from eth_tester import EthereumTester
from eth_tester.exceptions import TransactionFailed
from eth_typing.evm import HexAddress
from eth_utils import is_same_address
from eth_utils.units import units
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.types import ABI, Wei

from raiden_contracts.contract_manager import contracts_gas_path
from raiden_contracts.tests.utils import call_and_transact, get_random_privkey
from raiden_contracts.utils.logs import LogHandler
from raiden_contracts.utils.signature import private_key_to_address
from raiden_contracts.utils.type_aliases import PrivateKey


@pytest.fixture(scope="session")
def create_account(web3: Web3, ethereum_tester: EthereumTester) -> Callable:
    def get(privkey: Optional[PrivateKey] = None) -> HexAddress:
        if not privkey:
            privkey = get_random_privkey()
        address = private_key_to_address(privkey)

        if not any((is_same_address(address, x) for x in ethereum_tester.get_accounts())):
            # account has not been added to ethereum_tester, yet
            ethereum_tester.add_account(privkey.hex())

        for faucet in web3.eth.accounts[:10]:
            try:
                web3.eth.send_transaction(
                    {
                        "from": faucet,
                        "to": address,
                        "value": Wei(1 * int(units["finney"])),
                    }
                )
                break
            except TransactionFailed:
                continue
        return address

    return get


@pytest.fixture(scope="session")
def get_accounts(create_account: Callable) -> Callable:
    def get(number: int, privkeys: Iterable[PrivateKey] = ()) -> List:
        privkeys = iter(privkeys)
        return [create_account(privkey=next(privkeys, None)) for x in range(number)]

    return get


@pytest.fixture(scope="session")
def create_service_account(
    create_account: Callable, service_registry: Contract, custom_token: Contract
) -> Callable:
    """Returns an address registered to ServiceRegistry"""

    def get() -> HexAddress:
        account = create_account()
        deposit = service_registry.functions.currentPrice().call()
        call_and_transact(custom_token.functions.mint(deposit), {"from": account})
        call_and_transact(
            custom_token.functions.approve(service_registry.address, deposit),
            {"from": account},
        )
        call_and_transact(service_registry.functions.deposit(deposit), {"from": account})
        assert service_registry.functions.hasValidRegistration(account).call()
        return account

    return get


@pytest.fixture(scope="session")
def get_private_key(ethereum_tester: EthereumTester) -> Callable:
    def get(account_address: HexAddress) -> PrivateKey:
        keys = [
            key.to_bytes()
            for key in ethereum_tester.backend.account_keys
            if is_same_address(key.public_key.to_address(), account_address)
        ]
        assert len(keys) == 1
        return keys[0]

    return get


@pytest.fixture(scope="session")
def privkey_file(get_private_key: Callable, get_accounts: Callable) -> Iterable[IO]:
    (signer,) = get_accounts(1)
    priv_key = get_private_key(signer)
    with NamedTemporaryFile("w") as privkey_file:
        privkey_file.write(priv_key.hex())
        privkey_file.flush()
        yield privkey_file


@pytest.fixture(scope="session")
def event_handler(web3: Web3) -> Callable:
    def get(
        contract: Optional[Contract] = None,
        address: Optional[HexAddress] = None,
        abi: Optional[ABI] = None,
    ) -> LogHandler:
        if contract:
            abi = contract.abi
            address = contract.address

        if address and abi:
            return LogHandler(web3=web3, address=address, abi=abi)
        else:
            raise Exception("event_handler called without a contract instance")

    return get


@pytest.fixture
def txn_cost(web3: Web3, txn_gas: Callable) -> Callable:
    def get(txn_hash: str) -> int:
        return txn_gas(txn_hash) * web3.eth.gas_price

    return get


@pytest.fixture
def txn_gas(web3: Web3) -> Callable:
    def get(txn_hash: HexBytes) -> int:
        receipt = web3.eth.get_transaction_receipt(txn_hash)
        return receipt["gasUsed"]

    return get


@pytest.fixture(scope="session")
def gas_measurement_results() -> Dict:
    results: Dict = {}
    return results


def sys_args_contain(searched: str) -> bool:
    """Returns True if 'searched' appears in any of the command line arguments."""
    for arg in argv:
        if arg.find(searched) != -1:
            return True
    return False


@pytest.fixture
def print_gas(txn_gas: Callable, gas_measurement_results: Dict) -> Callable:
    def get(txn_hash: str, message: Optional[str] = None, additional_gas: int = 0) -> None:
        if not sys_args_contain("test_print_gas"):
            # If the command line arguments don't contain 'test_print_gas', do nothing
            return

        gas_used = txn_gas(txn_hash)
        if not message:
            message = txn_hash

        print("----------------------------------")
        print("GAS USED " + message, gas_used + additional_gas)
        print("----------------------------------")
        gas_measurement_results[message] = gas_used + additional_gas
        with contracts_gas_path().open(mode="w") as target_file:
            target_file.write(json.dumps(gas_measurement_results, sort_keys=True, indent=4))

    return get


@pytest.fixture()
def get_block(web3: Web3) -> Callable:
    def get(txn_hash: HexBytes) -> int:
        receipt = web3.eth.get_transaction_receipt(txn_hash)
        return receipt["blockNumber"]

    return get



import time
from copy import deepcopy
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

from eth_typing import ChecksumAddress, HexAddress
from eth_utils import encode_hex, is_address, to_checksum_address
from eth_utils.units import units
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract, ContractFunction
from web3.middleware import construct_sign_and_send_raw_middleware
from web3.types import ABI, TxParams, TxReceipt, Wei

from raiden_contracts.constants import (
    CONTRACT_MONITORING_SERVICE,
    CONTRACT_ONE_TO_N,
    CONTRACT_SECRET_REGISTRY,
    CONTRACT_SERVICE_REGISTRY,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    CONTRACT_USER_DEPOSIT,
    CONTRACTS_VERSION,
    DeploymentModule,
)
from raiden_contracts.contract_manager import (
    CompiledContract,
    DeployedContract,
    DeployedContracts,
    get_contracts_deployment_info,
)
from raiden_contracts.contract_source_manager import ContractSourceManager, contracts_source_path
from raiden_contracts.deploy.contract_verifier import ContractVerifier
from raiden_contracts.utils.file_ops import load_json_from_path
from raiden_contracts.utils.signature import private_key_to_address
from raiden_contracts.utils.transaction import check_successful_tx
from raiden_contracts.utils.type_aliases import ChainID, PrivateKey
from raiden_contracts.utils.versions import (
    contracts_version_monitoring_service_takes_token_network_registry,
)

LOG = getLogger(__name__)


class ContractDeployer(ContractVerifier):
    def __init__(
        self,
        web3: Web3,
        private_key: PrivateKey,
        gas_limit: int,
        gas_price: int,
        wait: int = 10,
        contracts_version: Optional[str] = None,
    ):
        # pylint: disable=E1101
        super(ContractDeployer, self).__init__(web3=web3, contracts_version=contracts_version)
        self.wait = wait
        self.owner = private_key_to_address(private_key)
        self.transaction = TxParams(
            {
                "from": self.owner,
                "gas": Wei(gas_limit),
                "gasPrice": Wei(gas_price * int(units["gwei"])),
            }
        )

        self.web3.middleware_onion.add(construct_sign_and_send_raw_middleware(private_key))

        # Check that the precompiled data matches the source code
        # Only for current version, because this is the only one with source code
        if self.contracts_version in [None, CONTRACTS_VERSION]:
            contract_manager_source = ContractSourceManager(
                contracts_source_path(contracts_version=contracts_version)
            )
            contract_manager_source.verify_precompiled_checksums(self.precompiled_path)
        else:
            LOG.info("Skipped checks against the source code because it is not available.")

    def deploy(self, contract_name: str, args: Optional[List] = None) -> TxReceipt:
        if args is None:
            args = list()
        contract_interface: CompiledContract = self.contract_manager.get_contract(contract_name)

        abi = contract_interface["abi"]
        bytecode = contract_interface["bin"]
        bytecode_runtime = None

        # Instantiate and deploy contract
        if bytecode_runtime:
            contract = self.web3.eth.contract(
                abi=abi, bytecode=bytecode, bytecode_runtime=bytecode_runtime
            )
        else:
            contract = self.web3.eth.contract(
                abi=abi,
                bytecode=bytecode,
            )

        # Get transaction hash from deployed contract
        txhash = self.send_deployment_transaction(contract=contract, args=args)

        # Get tx receipt to get contract address
        LOG.debug(
            f"Deploying {contract_name} txHash={encode_hex(txhash)}, "
            f"contracts version {self.contract_manager.contracts_version}"
        )
        receipt, tx = check_successful_tx(web3=self.web3, txid=txhash, timeout=self.wait)
        if not receipt["contractAddress"]:  # happens with Parity
            receipt["contractAddress"] = tx["creates"]  # type: ignore
        LOG.info(
            "{0} address: {1}. Gas used: {2}".format(
                contract_name, receipt["contractAddress"], receipt["gasUsed"]
            )
        )
        return receipt

    def transact(self, contract_method: ContractFunction) -> TxReceipt:
        """A wrapper around to_be_called.transact() that waits until the transaction succeeds."""
        txhash = contract_method.transact(self.transaction)
        LOG.debug(f"Sending txHash={encode_hex(txhash)}")
        receipt, _ = check_successful_tx(web3=self.web3, txid=txhash, timeout=self.wait)
        return receipt

    def send_deployment_transaction(self, contract: Type[Contract], args: List) -> HexBytes:
        txhash = None
        while txhash is None:
            try:
                txhash = contract.constructor(*args).transact(self.transaction)
            except ValueError as ex:
                # pylint: disable=E1126
                if ex.args[0]["code"] == -32015:
                    LOG.info(f"Deployment failed with {ex}. Retrying...")
                else:
                    raise ex

        return txhash

    def deploy_token_contract(
        self,
        token_supply: int,
        token_decimals: int,
        token_name: str,
        token_symbol: str,
        token_type: str = "CustomToken",
    ) -> Dict[str, ChecksumAddress]:
        """Deploy a token contract."""
        receipt = self.deploy(
            contract_name=token_type, args=[token_supply, token_decimals, token_name, token_symbol]
        )
        token_address = receipt["contractAddress"]
        assert token_address and is_address(token_address)
        token_address = to_checksum_address(token_address)
        return {token_type: token_address}

    def deploy_raiden_contracts(
        self,
        max_num_of_token_networks: int,
        reuse_secret_registry_from_deploy_file: Optional[Path],
        settle_timeout_min: int,
        settle_timeout_max: int,
    ) -> DeployedContracts:
        """Deploy all required raiden contracts and return a dict of contract_name:address

        Args:
            max_num_of_token_networks (Optional[int]): The max number of tokens that can be
            registered to the TokenNetworkRegistry. If None, the argument is omitted from
            the call to the constructor of TokenNetworkRegistry.
        """

        deployed_contracts: DeployedContracts = {
            "contracts_version": self.contract_manager.contracts_version,
            "chain_id": self.web3.eth.chain_id,
            "contracts": {},
        }

        if reuse_secret_registry_from_deploy_file:
            reused_doc = DeployedContracts(  # type: ignore
                load_json_from_path(reuse_secret_registry_from_deploy_file)
            )
            if not reused_doc:
                raise RuntimeError(
                    f"{reuse_secret_registry_from_deploy_file} does not contain deployment data."
                )
            reused_contracts = reused_doc["contracts"]

            secret_registry = self.contract_instance_from_deployment_data(
                reused_doc, CONTRACT_SECRET_REGISTRY
            )

            deployed_contracts["contracts"][CONTRACT_SECRET_REGISTRY] = deepcopy(
                reused_contracts[CONTRACT_SECRET_REGISTRY]
            )
        else:
            secret_registry = self._deploy_and_remember(
                contract_name=CONTRACT_SECRET_REGISTRY,
                arguments=[],
                deployed_contracts=deployed_contracts,
            )

        token_network_registry_args = [
            secret_registry.address,
            deployed_contracts["chain_id"],
            settle_timeout_min,
            settle_timeout_max,
            max_num_of_token_networks,
        ]
        self._deploy_and_remember(
            contract_name=CONTRACT_TOKEN_NETWORK_REGISTRY,
            arguments=token_network_registry_args,
            deployed_contracts=deployed_contracts,
        )

        return deployed_contracts

    def _deploy_and_remember(
        self,
        contract_name: str,
        arguments: List,
        deployed_contracts: DeployedContracts,
    ) -> Contract:
        """Deploys contract_name with arguments and store the result in deployed_contracts."""
        receipt = self.deploy(contract_name, arguments)
        deployed_contracts["contracts"][contract_name] = _deployed_data_from_receipt(
            receipt=receipt, constructor_arguments=arguments
        )
        return self.web3.eth.contract(
            abi=self.contract_manager.get_contract_abi(contract_name),
            address=deployed_contracts["contracts"][contract_name]["address"],
        )

    def register_token_network(
        self,
        token_registry_abi: ABI,
        token_registry_address: ChecksumAddress,
        token_address: ChecksumAddress,
        channel_participant_deposit_limit: int,
        token_network_deposit_limit: int,
    ) -> Dict[str, Any]:
        """Register token with a TokenNetworkRegistry contract."""
        token_network_registry = self.web3.eth.contract(
            abi=token_registry_abi, address=token_registry_address
        )

        command = token_network_registry.functions.createERC20TokenNetwork(
            token_address, channel_participant_deposit_limit, token_network_deposit_limit
        )
        self.transact(command)

        LOG.debug("Collecting constructor parameters for later verification")
        chain_id = ChainID(self.web3.eth.chain_id)
        deployment_data = get_contracts_deployment_info(
            chain_id=chain_id,
            version=self.contract_manager.contracts_version,
            module=DeploymentModule.RAIDEN,
        )
        assert deployment_data
        secret_registry = self.contract_instance_from_deployment_data(
            deployment_data, CONTRACT_SECRET_REGISTRY
        )
        token_network_registry = self.contract_instance_from_deployment_data(
            deployment_data, CONTRACT_TOKEN_NETWORK_REGISTRY
        )
        settle_timeout_min = token_network_registry.functions.settlement_timeout_min().call()
        settle_timeout_max = token_network_registry.functions.settlement_timeout_max().call()
        constructor_arguments = dict(
            _token_address=token_address,
            _secret_registry=secret_registry.address,
            _chain_id=chain_id,
            _settlement_timeout_min=settle_timeout_min,
            _settlement_timeout_max=settle_timeout_max,
            _controller=self.owner,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

        LOG.debug("Getting address of new token network")
        token_network_address = "0x0"
        while int(token_network_address, 16) == 0:
            token_network_address = token_network_registry.functions.token_to_token_networks(
                token_address
            ).call()
            time.sleep(2)
        LOG.debug(f"TokenNetwork address: {token_network_address}")
        return dict(
            token_network_address=token_network_address,
            constructor_arguments=constructor_arguments,
        )

    def deploy_service_contracts(
        self,
        token_address: HexAddress,
        user_deposit_whole_balance_limit: int,
        service_registry_controller: HexAddress,
        initial_service_deposit_price: int,
        service_deposit_bump_numerator: int,
        service_deposit_bump_denominator: int,
        decay_constant: int,
        min_price: int,
        registration_duration: int,
        token_network_registry_address: HexAddress,
        reuse_service_registry_from_deploy_file: Optional[Path],
    ) -> DeployedContracts:
        """Deploy 3rd party service contracts"""
        if not contracts_version_monitoring_service_takes_token_network_registry(
            self.contract_manager.contracts_version
        ):
            raise RuntimeError("Deployment of older service contracts is not suppported.")

        chain_id = self.web3.eth.chain_id
        deployed_contracts: DeployedContracts = {
            "contracts_version": self.contract_manager.contracts_version,
            "chain_id": chain_id,
            "contracts": {},
        }

        if reuse_service_registry_from_deploy_file:
            reused_doc = DeployedContracts(  # type: ignore
                load_json_from_path(reuse_service_registry_from_deploy_file)
            )
            if not reused_doc:
                raise RuntimeError(
                    f"{reuse_service_registry_from_deploy_file} does not contain deployment data."
                )
            reused_contracts = reused_doc["contracts"]

            service_registry = self.contract_instance_from_deployment_data(
                reused_doc, CONTRACT_SERVICE_REGISTRY
            )

            deployed_contracts["contracts"][CONTRACT_SERVICE_REGISTRY] = deepcopy(
                reused_contracts[CONTRACT_SERVICE_REGISTRY]
            )
        else:
            service_registry = self._deploy_and_remember(
                contract_name=CONTRACT_SERVICE_REGISTRY,
                arguments=[
                    token_address,
                    service_registry_controller,
                    initial_service_deposit_price,
                    service_deposit_bump_numerator,
                    service_deposit_bump_denominator,
                    decay_constant,
                    min_price,
                    registration_duration,
                ],
                deployed_contracts=deployed_contracts,
            )

        user_deposit = self._deploy_and_remember(
            contract_name=CONTRACT_USER_DEPOSIT,
            arguments=[token_address, user_deposit_whole_balance_limit],
            deployed_contracts=deployed_contracts,
        )

        monitoring_service_constructor_args = [
            token_address,
            deployed_contracts["contracts"][CONTRACT_SERVICE_REGISTRY]["address"],
            deployed_contracts["contracts"][CONTRACT_USER_DEPOSIT]["address"],
            token_network_registry_address,
        ]
        msc = self._deploy_and_remember(
            contract_name=CONTRACT_MONITORING_SERVICE,
            arguments=monitoring_service_constructor_args,
            deployed_contracts=deployed_contracts,
        )

        one_to_n = self._deploy_and_remember(
            contract_name=CONTRACT_ONE_TO_N,
            arguments=[user_deposit.address, chain_id, service_registry.address],
            deployed_contracts=deployed_contracts,
        )

        # Tell the UserDeposit instance about other contracts.
        LOG.debug(
            "Calling UserDeposit.init() with "
            f"msc_address={msc.address} "
            f"one_to_n_address={one_to_n.address}"
        )
        self.transact(
            user_deposit.functions.init(
                _msc_address=msc.address, _one_to_n_address=one_to_n.address
            )
        )

        return deployed_contracts


def _deployed_data_from_receipt(
    receipt: TxReceipt, constructor_arguments: List
) -> DeployedContract:
    assert receipt["contractAddress"] is not None
    return {
        "address": to_checksum_address(receipt["contractAddress"]),
        "transaction_hash": encode_hex(receipt["transactionHash"]),
        "block_number": receipt["blockNumber"],
        "gas_cost": receipt["gasUsed"],
        "constructor_arguments": constructor_arguments,
    }



from typing import Any, Callable, Dict, Tuple

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MIN, ParticipantInfoIndex
from raiden_contracts.tests.fixtures.channel import call_settle
from raiden_contracts.tests.fixtures.channel_test_values import (
    channel_settle_invalid_test_values,
    channel_settle_test_values,
)
from raiden_contracts.tests.utils import (
    NONEXISTENT_LOCKSROOT,
    UINT256_MAX,
    are_balance_proofs_valid,
    call_and_transact,
    get_expected_after_settlement_unlock_amounts,
    get_onchain_settlement_amounts,
    get_settlement_amounts,
    get_total_available_deposit,
    get_unlocked_amount,
    is_balance_proof_old,
    were_balance_proofs_valid,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.pending_transfers import (
    get_pending_transfers_tree_with_generated_lists,
)


@pytest.fixture()
def test_settlement_outcome(
    web3: Web3,
    secret_registry_contract: Contract,
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    close_and_update_channel: Callable,
    settle_state_tests: Callable,
    reveal_secrets: Callable,
) -> Callable:
    def f(
        participants: Tuple,
        channel_values: Tuple,
        expected_final_balance_A0: int,
        expected_final_balance_B0: int,
    ) -> None:
        (A, B) = participants
        (vals_A, vals_B, balance_proof_type) = channel_values
        assert were_balance_proofs_valid(vals_A, vals_B)

        # Start channel lifecycle
        channel_identifier = create_channel_and_deposit(A, B, vals_A.deposit, vals_B.deposit)
        withdraw_channel(channel_identifier, A, vals_A.withdrawn, UINT256_MAX, B)
        withdraw_channel(channel_identifier, B, vals_B.withdrawn, UINT256_MAX, A)

        # For the purpose of this test, it is not important when the secrets are revealed,
        # as long as the secrets connected to pending transfers that should be finalized,
        # are revealed before their expiration.

        # Mock pending transfers data for A -> B
        pending_transfers_tree_A = get_pending_transfers_tree_with_generated_lists(
            web3,
            unlockable_amount=vals_A.locked_amounts.claimable_locked,
            expired_amount=vals_A.locked_amounts.unclaimable_locked,
        )
        vals_A.locksroot = pending_transfers_tree_A.hash_of_packed_transfers
        # Reveal A's secrets.
        reveal_secrets(A, pending_transfers_tree_A.unlockable)

        # Mock pending transfers data for B -> A
        pending_transfers_tree_B = get_pending_transfers_tree_with_generated_lists(
            web3,
            unlockable_amount=vals_B.locked_amounts.claimable_locked,
            expired_amount=vals_B.locked_amounts.unclaimable_locked,
        )
        vals_B.locksroot = pending_transfers_tree_B.hash_of_packed_transfers
        # Reveal B's secrets
        reveal_secrets(B, pending_transfers_tree_B.unlockable)

        close_and_update_channel(channel_identifier, A, vals_A, B, vals_B)

        mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN)

        pre_balance_A = custom_token.functions.balanceOf(A).call()
        pre_balance_B = custom_token.functions.balanceOf(B).call()
        pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

        # We do the balance & state tests here for each channel and also compare with
        # the expected settlement amounts
        settle_state_tests(
            channel_identifier,
            A,
            vals_A,
            B,
            vals_B,
            pre_balance_A,
            pre_balance_B,
            pre_balance_contract,
        )

        # We compute again the settlement amounts here to compare with the other channel
        # settlement test values, which should be equal

        # Calculate how much A and B should receive
        settlement_equivalent = get_settlement_amounts(vals_A, vals_B)

        # Calculate how much A and B receive according to onchain computation
        settlement_onchain_equivalent = get_onchain_settlement_amounts(vals_A, vals_B)

        assert (
            settlement_equivalent.participant1_balance
            == settlement_onchain_equivalent.participant1_balance
        )
        assert (
            settlement_equivalent.participant2_balance
            == settlement_onchain_equivalent.participant2_balance
        )
        assert (
            settlement_equivalent.participant1_locked
            == settlement_onchain_equivalent.participant1_locked
        )
        assert (
            settlement_equivalent.participant2_locked
            == settlement_onchain_equivalent.participant2_locked
        )

        assert (
            get_unlocked_amount(
                secret_registry_contract, pending_transfers_tree_B.packed_transfers
            )
            == vals_B.locked_amounts.claimable_locked
        )

        # A unlocks B's pending transfers
        info_B = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
        assert (
            settlement_equivalent.participant2_locked == info_B[ParticipantInfoIndex.LOCKED_AMOUNT]
        )

        if info_B[ParticipantInfoIndex.LOCKED_AMOUNT] == 0:
            with pytest.raises(TransactionFailed):
                token_network.functions.unlock(
                    channel_identifier, A, B, pending_transfers_tree_B.packed_transfers
                ).call()
        else:
            call_and_transact(
                token_network.functions.unlock(
                    channel_identifier, A, B, pending_transfers_tree_B.packed_transfers
                )
            )

        # The locked amount should have been removed from contract storage
        info_B = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
        assert info_B[ParticipantInfoIndex.LOCKED_AMOUNT] == 0
        assert info_B[ParticipantInfoIndex.LOCKSROOT] == NONEXISTENT_LOCKSROOT

        # B unlocks A's pending transfers
        info_A = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
        assert (
            settlement_equivalent.participant1_locked == info_A[ParticipantInfoIndex.LOCKED_AMOUNT]
        )

        if info_A[ParticipantInfoIndex.LOCKED_AMOUNT] == 0:
            with pytest.raises(TransactionFailed):
                token_network.functions.unlock(
                    channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
                ).call()
        else:
            call_and_transact(
                token_network.functions.unlock(
                    channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
                )
            )

        # The locked amount should have been removed from contract storage
        info_A = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
        assert info_A[ParticipantInfoIndex.LOCKED_AMOUNT] == 0
        assert info_A[ParticipantInfoIndex.LOCKSROOT] == NONEXISTENT_LOCKSROOT

        # Do the post settlement and unlock tests for valid balance proofs
        balance_A = custom_token.functions.balanceOf(A).call()
        balance_B = custom_token.functions.balanceOf(B).call()

        # We MUST ensure balance correctness for valid last balance proofs
        if balance_proof_type == "valid":
            # Calculate how much A and B should receive after the channel is settled and
            # unlock is called by both.
            (
                expected_final_balance_A,
                expected_final_balance_B,
            ) = get_expected_after_settlement_unlock_amounts(vals_A, vals_B)
            expected_balance_A = pre_balance_A + expected_final_balance_A
            expected_balance_B = pre_balance_B + expected_final_balance_B

            assert balance_A == expected_balance_A
            assert balance_B <= expected_balance_B

        # For balance proofs where one of them is old, we need to compare with the expected
        # final balances for the valid last balance proofs
        expected_balance_A = pre_balance_A + expected_final_balance_A0
        expected_balance_B = pre_balance_B + expected_final_balance_B0

        # Tests for when B has submitted an old balance proof for A
        # A must not receive less tokens than expected with a valid last balance proof
        # B must not receive more tokens than expected with a valid last balance proof
        if balance_proof_type == "old_last":
            assert balance_A >= expected_balance_A
            assert balance_B <= expected_balance_B

        # Tests for when A has submitted an old balance proof for B
        # A must not receive more tokens than expected with a valid last balance proof
        # B must not receive less tokens than expected with a valid last balance proof
        if balance_proof_type == "last_old":
            assert balance_A <= expected_balance_A
            assert balance_B >= expected_balance_B

        # Regardless of the tokens received by the two participants, we must make sure tokens
        # are not stolen from the other channels. And we must make sure tokens are not locked in
        # the contract after the entire channel cycle is finalized.
        final_contract_balance = custom_token.functions.balanceOf(token_network.address).call()
        assert final_contract_balance == pre_balance_contract - get_total_available_deposit(
            vals_A, vals_B
        )
        assert custom_token.functions.balanceOf(token_network.address).call() == (
            pre_balance_contract - (expected_final_balance_A0 + expected_final_balance_B0)
        )

    return f


@pytest.mark.slow
@pytest.mark.parametrize("channel_test_values", channel_settle_test_values)
@pytest.mark.parametrize("tested_range", ("one_old", "both_old_1", "both_old_2"))
# This test is split in three so it does not time out on travis
def test_channel_settle_old_balance_proof_values(
    get_accounts: Callable,
    assign_tokens: Callable,
    channel_test_values: Dict[str, Any],
    create_channel_and_deposit: Callable,
    test_settlement_outcome: Callable,
    tested_range: str,
) -> None:
    """Test the settlement implementation when both/one of the balance proofs are outdated"""
    (A, B, C, D) = get_accounts(4)
    (vals_A0, vals_B0) = channel_test_values["valid_last"]
    assert are_balance_proofs_valid(vals_A0, vals_B0)
    assert not is_balance_proof_old(vals_A0, vals_B0)

    # Mint additional tokens for participants
    assign_tokens(A, 400)
    assign_tokens(B, 200)
    # We make sure the contract has more tokens than A, B will deposit
    create_channel_and_deposit(C, D, 40, 60)

    # Calculate the final expected balances after the channel lifecycle, (after settlement and
    # unlocks), when we know how the locked amounts will be distributed.
    # This is a very important check. This must be true for the last known balance proofs,
    # but also for a last known balance proof + an old balance proof provided on purpose or not.
    # Where a valid balance proof is a balance proof that respects the Raiden client
    # value contraints, as defined here:
    # https://github.com/raiden-network/raiden-contracts/issues/188#issuecomment-404752095
    (
        expected_final_balance_A0,
        expected_final_balance_B0,
    ) = get_expected_after_settlement_unlock_amounts(vals_A0, vals_B0)

    if tested_range == "one_old":

        test_settlement_outcome(
            (A, B),
            (vals_A0, vals_B0, "valid"),
            expected_final_balance_A0,
            expected_final_balance_B0,
        )

        if "old_last" in channel_test_values:
            for vals_A in channel_test_values["old_last"]:
                vals_B = vals_B0
                test_settlement_outcome(
                    (A, B),
                    (vals_A, vals_B, "old_last"),
                    expected_final_balance_A0,
                    expected_final_balance_B0,
                )

        if "last_old" in channel_test_values:
            for vals_B in channel_test_values["last_old"]:
                vals_A = vals_A0
                test_settlement_outcome(
                    (A, B),
                    (vals_A, vals_B, "last_old"),
                    expected_final_balance_A0,
                    expected_final_balance_B0,
                )

    elif tested_range == "both_old_1":

        if "old_last" in channel_test_values and "last_old" in channel_test_values:
            for vals_A in channel_test_values["old_last"][:5]:
                for vals_B in channel_test_values["last_old"][:5]:
                    # We only need to test for cases  where the we have the same argument ordering
                    # for settleChannel, keeping the order of balance calculations
                    B_total = vals_B.transferred + vals_B.locked_amounts.locked
                    A_total = vals_A.transferred + vals_A.locked_amounts.locked
                    if B_total >= A_total:
                        test_settlement_outcome(
                            (A, B),
                            (vals_A, vals_B, "invalid"),
                            expected_final_balance_A0,
                            expected_final_balance_B0,
                        )

    else:

        if "old_last" in channel_test_values and "last_old" in channel_test_values:
            for vals_A in channel_test_values["old_last"][5:]:
                for vals_B in channel_test_values["last_old"][5:]:
                    # We only need to test for cases  where the we have the same argument ordering
                    # for settleChannel, keeping the order of balance calculations
                    B_total = vals_B.transferred + vals_B.locked_amounts.locked
                    A_total = vals_A.transferred + vals_A.locked_amounts.locked
                    if B_total >= A_total:
                        test_settlement_outcome(
                            (A, B),
                            (vals_A, vals_B, "invalid"),
                            expected_final_balance_A0,
                            expected_final_balance_B0,
                        )


@pytest.mark.slow
@pytest.mark.parametrize("channel_test_values", channel_settle_invalid_test_values)
def test_channel_settle_invalid_balance_proof_values(
    web3: Web3,
    get_accounts: Callable,
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    close_and_update_channel: Callable,
    settle_state_tests: Callable,
    reveal_secrets: Callable,
    channel_test_values: Tuple,
) -> None:
    """Check the settlement results with invalid balance proofs"""
    (A, B, C, D) = get_accounts(4)
    (vals_A, vals_B) = channel_test_values

    # We just need to test that settleChannel does not fail
    # We cannot ensure correctly computed final balances if the balance proofs
    # are invalid.
    # We can just test that participants do not get more tokens than the channel deposit

    # We make sure the contract has more tokens than A, B will deposit
    create_channel_and_deposit(C, D, 40, 60)

    # Start channel lifecycle for A, B
    channel_identifier = create_channel_and_deposit(A, B, vals_A.deposit, vals_B.deposit)
    withdraw_channel(channel_identifier, A, vals_A.withdrawn, UINT256_MAX, B)
    withdraw_channel(channel_identifier, B, vals_B.withdrawn, UINT256_MAX, A)

    # For the purpose of this test, it is not important when the secrets are revealed,
    # as long as the secrets connected to pending transfers that should be finalized,
    # are revealed before their expiration.

    # Mock pending transfers data for A -> B
    pending_transfers_tree_A = get_pending_transfers_tree_with_generated_lists(
        web3=web3,
        unlockable_amount=vals_A.locked_amounts.claimable_locked,
        expired_amount=vals_A.locked_amounts.unclaimable_locked,
    )
    vals_A.locksroot = pending_transfers_tree_A.hash_of_packed_transfers
    # Reveal A's secrets.
    reveal_secrets(A, pending_transfers_tree_A.unlockable)

    # Mock pending transfers data for B -> A
    pending_transfers_tree_B = get_pending_transfers_tree_with_generated_lists(
        web3,
        unlockable_amount=vals_B.locked_amounts.claimable_locked,
        expired_amount=vals_B.locked_amounts.unclaimable_locked,
    )
    vals_B.locksroot = pending_transfers_tree_B.hash_of_packed_transfers
    # Reveal B's secrets
    reveal_secrets(B, pending_transfers_tree_B.unlockable)

    close_and_update_channel(channel_identifier, A, vals_A, B, vals_B)

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    # We do the balance & state tests here for each channel and also compare with
    # the expected settlement amounts
    settle_state_tests(
        channel_identifier,
        A,
        vals_A,
        B,
        vals_B,
        pre_balance_A,
        pre_balance_B,
        pre_balance_contract,
    )



import json
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, Optional


def load_json_from_path(f: Path) -> Optional[Dict[str, Any]]:
    try:
        with f.open() as deployment_file:
            return json.load(deployment_file)
    except FileNotFoundError:
        return None
    except (JSONDecodeError, UnicodeDecodeError) as ex:
        raise ValueError(f"Deployment data file is corrupted: {ex}") from ex



from typing import NewType

T_AdditionalHash = bytes
AdditionalHash = NewType("AdditionalHash", T_AdditionalHash)

T_BalanceHash = bytes
BalanceHash = NewType("BalanceHash", T_BalanceHash)

# An absolute number of blocks
T_BlockExpiration = int
BlockExpiration = NewType("BlockExpiration", T_BlockExpiration)

T_ChainID = int
ChainID = NewType("ChainID", T_ChainID)

T_ChannelID = int
ChannelID = NewType("ChannelID", T_ChannelID)

T_Locksroot = bytes
Locksroot = NewType("Locksroot", T_Locksroot)

T_PrivateKey = bytes
PrivateKey = NewType("PrivateKey", T_PrivateKey)

T_Signature = bytes
Signature = NewType("Signature", T_Signature)

T_TokenAmount = int
TokenAmount = NewType("TokenAmount", T_TokenAmount)



from typing import Callable

import pytest
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MAX, TEST_SETTLE_TIMEOUT_MIN


@pytest.fixture
def token_network_test_storage(
    deploy_tester_contract: Callable,
    web3: Web3,
    custom_token: Contract,
    secret_registry_contract: Contract,
) -> Contract:
    return deploy_tester_contract(
        "TokenNetworkInternalStorageTest",
        [
            custom_token.address,
            secret_registry_contract.address,
            web3.eth.chain_id,
            TEST_SETTLE_TIMEOUT_MIN,
            TEST_SETTLE_TIMEOUT_MAX,
        ],
    )


@pytest.fixture
def token_network_test_signatures(
    deploy_tester_contract: Callable,
    web3: Web3,
    custom_token: Contract,
    secret_registry_contract: Contract,
) -> Contract:
    return deploy_tester_contract(
        "TokenNetworkSignatureTest",
        _token_address=custom_token.address,
        _secret_registry=secret_registry_contract.address,
        _chain_id=web3.eth.chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
    )


@pytest.fixture
def token_network_test_utils(
    deploy_tester_contract: Callable,
    web3: Web3,
    custom_token: Contract,
    secret_registry_contract: Contract,
) -> Contract:
    return deploy_tester_contract(
        "TokenNetworkUtilsTest",
        _token_address=custom_token.address,
        _secret_registry=secret_registry_contract.address,
        _chain_id=web3.eth.chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
    )


@pytest.fixture
def signature_test_contract(deploy_tester_contract: Callable) -> Contract:
    return deploy_tester_contract("SignatureVerifyTest")



import logging
from typing import Any, Callable, Tuple

import pytest
from eth_tester.exceptions import TransactionFailed
from eth_typing import HexAddress
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.tests.utils.constants import DEPLOYER_ADDRESS

log = logging.getLogger(__name__)


def deploy_contract_txhash(
    web3: Web3,
    contracts_manager: ContractManager,
    deployer_address: HexAddress,
    contract_name: str,
    **kwargs: Any,
) -> Tuple[HexAddress, Contract]:
    json_contract = contracts_manager.get_contract(contract_name)
    abi = json_contract["abi"]
    bytecode = json_contract["bin"]
    bytecode_runtime = None

    if bytecode_runtime is not None:
        contract = web3.eth.contract(abi=abi, bytecode=bytecode, bytecode_runtime=bytecode_runtime)
    else:
        contract = web3.eth.contract(abi=abi, bytecode=bytecode)

    mine_blocks(web3, 3)
    # Failure does not fire an exception. Check the receipt for status.
    txhash = contract.constructor(**kwargs).transact({"from": deployer_address})
    mine_blocks(web3, 1)

    receipt = web3.eth.get_transaction_receipt(txhash)
    if receipt["status"] != 1:
        raise TransactionFailed("deployment failed")

    return txhash, contract(receipt["contractAddress"])


@pytest.fixture
def deploy_tester_contract_txhash(web3: Web3, contracts_manager: ContractManager) -> Callable:
    """Returns a function that can be used to deploy a named contract,
    but returning txhash only"""

    def f(
        contract_name: str,
        deployer_address: HexAddress = DEPLOYER_ADDRESS,
        **kwargs: Any,
    ) -> HexAddress:
        txhash, _ = deploy_contract_txhash(
            web3,
            contracts_manager,
            deployer_address,
            contract_name,
            **kwargs,
        )
        return txhash

    return f


@pytest.fixture(scope="session")
def deploy_tester_contract(web3: Web3, contracts_manager: ContractManager) -> Callable:
    """Returns a function that can be used to deploy a named contract,
    using contract manager to compile the bytecode and get the ABI"""

    def f(
        contract_name: str,
        deployer_address: HexAddress = DEPLOYER_ADDRESS,
        **kwargs: Any,
    ) -> Contract:
        _, contract = deploy_contract_txhash(
            web3,
            contracts_manager,
            deployer_address,
            contract_name,
            **kwargs,
        )
        return contract

    return f



"""ContractManager knows binaries and ABI of contracts."""
import enum
import json
from copy import deepcopy
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Optional

from eth_typing import HexStr
from eth_typing.evm import ChecksumAddress
from mypy_extensions import TypedDict
from web3.types import ABI, ABIEvent

from raiden_contracts.constants import ID_TO_CHAINNAME, DeploymentModule
from raiden_contracts.utils.file_ops import load_json_from_path
from raiden_contracts.utils.type_aliases import ChainID
from raiden_contracts.utils.versions import contracts_version_provides_services

_BASE = Path(__file__).parent


# Classes for static type checking of deployed_contracts dictionary.


CompiledContract = TypedDict(
    "CompiledContract",
    {"abi": ABI, "bin-runtime": str, "bin": str, "metadata": str},
)


DeployedContract = TypedDict(
    "DeployedContract",
    {
        "address": ChecksumAddress,
        "transaction_hash": HexStr,
        "block_number": int,
        "gas_cost": int,
        "constructor_arguments": List[Any],
    },
)


class DeployedContracts(TypedDict, total=False):
    chain_id: int
    contracts: Dict[str, DeployedContract]
    contracts_version: str
    token_networks: List[Dict[str, Any]]


class ContractManagerLoadError(RuntimeError):
    """Failure in loading contracts.json."""


class ContractManager:
    """ContractManager holds compiled contracts of the same version

    Provides access to the ABI and the bytecode.
    """

    def __init__(self, path: Path) -> None:
        """Params:
        path: path to a precompiled contract JSON file,
        """
        try:
            with path.open() as precompiled_file:
                precompiled_content = json.load(precompiled_file)
        except (JSONDecodeError, UnicodeDecodeError) as ex:
            raise ContractManagerLoadError(f"Can't load precompiled smart contracts: {ex}") from ex
        try:
            self.contracts: Dict[str, CompiledContract] = precompiled_content["contracts"]
            if not self.contracts:
                raise RuntimeError(
                    f"Cannot find precompiled contracts data in the JSON file {path}."
                )
            self.overall_checksum = precompiled_content["overall_checksum"]
            self.contracts_checksums = precompiled_content["contracts_checksums"]
            self.contracts_version = precompiled_content["contracts_version"]
        except KeyError as ex:
            raise ContractManagerLoadError(
                f"Precompiled contracts json has unexpected format: {ex}"
            ) from ex

    def get_contract(self, contract_name: str) -> CompiledContract:
        """Return ABI, BIN of the given contract."""
        assert self.contracts, "ContractManager should have contracts compiled"
        try:
            return self.contracts[contract_name]
        except KeyError:
            raise KeyError(
                f"contracts_version {self.contracts_version} does not have {contract_name}"
            )

    def has_contract(self, contract_name: str) -> bool:
        return contract_name in self.contracts

    def get_contract_abi(self, contract_name: str) -> ABI:
        """Returns the ABI for a given contract."""
        assert self.contracts, "ContractManager should have contracts compiled"
        return self.contracts[contract_name]["abi"]

    def get_event_abi(self, contract_name: str, event_name: str) -> ABIEvent:
        """Returns the ABI for a given event."""
        # Import locally to avoid web3 dependency during installation via `compile_contracts`
        from web3._utils.contracts import find_matching_event_abi

        assert self.contracts, "ContractManager should have contracts compiled"
        contract_abi = self.get_contract_abi(contract_name)
        return find_matching_event_abi(abi=contract_abi, event_name=event_name)

    def get_constructor_argument_types(self, contract_name: str) -> List:
        abi = self.get_contract_abi(contract_name=contract_name)
        constructor = [f for f in abi if f["type"] == "constructor"][0]
        return [arg["type"] for arg in constructor["inputs"]]

    def get_runtime_hexcode(self, contract_name: str) -> str:
        """Calculate the runtime hexcode with 0x prefix.

        Parameters:
            contract_name: name of the contract such as CONTRACT_TOKEN_NETWORK
        """
        return "0x" + self.contracts[contract_name]["bin-runtime"]


def contracts_data_path(version: Optional[str] = None) -> Path:
    """Returns the deployment data directory for a version."""
    if version is None:
        return _BASE.joinpath("data")
    return _BASE.joinpath(f"data_{version}")


def contracts_precompiled_path(version: Optional[str] = None) -> Path:
    """Returns the path of JSON file where the bytecode can be found."""
    data_path = contracts_data_path(version)
    return data_path.joinpath("contracts.json")


def contracts_gas_path(version: Optional[str] = None) -> Any:
    """Returns the path of JSON file where the gas usage information can be found."""
    data_path = contracts_data_path(version)
    return data_path.joinpath("gas.json")


def gas_measurements(version: Optional[str] = None) -> Dict[str, int]:
    """Returns gas usage measurement."""
    json_path = contracts_gas_path(version)
    with json_path.open() as gas_file:
        return json.load(gas_file)


class ContractDevEnvironment(enum.Enum):
    DEMO = "demo"
    UNSTABLE = "unstable"


def contracts_deployed_path(
    chain_id: ChainID,
    version: Optional[str] = None,
    services: bool = False,
    development_environment: ContractDevEnvironment = ContractDevEnvironment.DEMO,
) -> Path:
    """Returns the path of the deplolyment data JSON file."""
    data_path = contracts_data_path(version)
    chain_name = ID_TO_CHAINNAME[chain_id] if chain_id in ID_TO_CHAINNAME else "private_net"

    if development_environment == ContractDevEnvironment.UNSTABLE:
        assert chain_name == "goerli", "Unstable contracts only available on goerli"
        chain_name = "goerli_unstable"

    return data_path.joinpath(f'deployment_{"services_" if services else ""}{chain_name}.json')


def merge_deployment_data(dict1: DeployedContracts, dict2: DeployedContracts) -> DeployedContracts:
    """Take contents of two deployment JSON files and merge them

    The dictionary under 'contracts' key will be merged. The 'contracts'
    contents from different JSON files must not overlap. The contents
    under other keys must be identical.
    """
    if not dict1:
        return dict2
    if not dict2:
        return dict1
    common_contracts: Dict[str, DeployedContract] = deepcopy(dict1["contracts"])
    if common_contracts.keys() & dict2["contracts"].keys():
        raise ValueError(
            "dict1 and dict2 contain contracts of the same name. "
            "Now failing instead of overwriting any of these."
        )
    common_contracts.update(dict2["contracts"])

    if dict2["chain_id"] != dict1["chain_id"]:
        raise ValueError("Got dictionaries with different chain_id's.")
    if dict2["contracts_version"] != dict1["contracts_version"]:
        raise ValueError("Got dictionaries with different contracts_versions.")

    return DeployedContracts(
        {
            "contracts": common_contracts,
            "chain_id": dict1["chain_id"],
            "contracts_version": dict1["contracts_version"],
        }
    )


def get_contracts_deployment_info(
    chain_id: ChainID,
    version: Optional[str] = None,
    module: DeploymentModule = DeploymentModule.ALL,
    development_environment: ContractDevEnvironment = ContractDevEnvironment.DEMO,
) -> Optional[DeployedContracts]:
    """Reads the deployment data. Returns None if the file is not found.

    Parameter:
        module The name of the module. ALL means deployed contracts from all modules that are
        available for the version.
    """
    if not isinstance(module, DeploymentModule):
        raise ValueError(f"Unknown module {module} given to get_contracts_deployment_info()")

    def module_chosen(to_be_added: DeploymentModule) -> bool:
        return module == to_be_added or module == DeploymentModule.ALL

    files: List[Path] = []

    if module_chosen(DeploymentModule.RAIDEN):
        files.append(
            contracts_deployed_path(
                chain_id=chain_id,
                version=version,
                services=False,
                development_environment=development_environment,
            )
        )

    if module == DeploymentModule.SERVICES and not contracts_version_provides_services(version):
        raise ValueError(
            f"SERVICES module queried for version {version}, but {version} "
            "does not provide service contracts."
        )

    if module_chosen(DeploymentModule.SERVICES) and contracts_version_provides_services(version):
        files.append(
            contracts_deployed_path(
                chain_id=chain_id,
                version=version,
                services=True,
                development_environment=development_environment,
            )
        )
    deployment_data: Optional[DeployedContracts] = {}

    for f in files:
        j = load_json_from_path(f)
        if j is None:
            continue
        assert deployment_data is not None
        deployment_data = merge_deployment_data(
            deployment_data,
            DeployedContracts(
                {
                    "chain_id": j["chain_id"],
                    "contracts": j["contracts"],
                    "contracts_version": j["contracts_version"],
                }
            ),
        )

    if not deployment_data:
        deployment_data = None
    return deployment_data
